//---------------------------------------------------------------------------

#include "tnlib.h"
#pragma hdrstop
#include <shlwapi.h>

#include "FindFileTask.h"
#include "filestr.h"
#include "ffstr.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)

//#pragma lnk shlwapi.lib

bool CompareMask(tnstr_vec &masks, const tchar *filename);

void TFindFileTask::Setup(const tchar *pattern, const tchar *dir, const tchar *masks, int option, FNFoundCallback foundcallback, void *userdata)
{
	Pattern = pattern;
	Dir = dir;
	if (masks){
		Masks.split(masks, _t(";"));
	}
	Option = option;
	FoundCallback = foundcallback;
	UserData = userdata;
}
void TFindFileTask::Stop()
{
	FoundCallback = NULL;
	super::Stop();
	if (FindStrTask){
		FindStrTask->StopCallback();
	}
}

void TFindFileTask::Execute()
{
	FindStrTask = new TFindStrTask();
	FindStrTask->Setup(Pattern, Option, FoundCallback, UserData);
	FindStrTask->Start();

	FindDirs(Dir);

	if (FindStrTask){
		if (StopReq){
			// Cancel/Stop. FindStrTask will destroy itself.
			FindStrTask->Stop();
			FindStrTask = NULL;
		} else {
			// Normal termination. Wait for finishing all files.
			FindStrTask->FreeOnTerminate(false);
			FindStrTask->Done();
			FindStrTask->WaitFor();
			Lock();
			delete FindStrTask;
			FindStrTask = NULL;
			Unlock();
		}
	}

	if (FoundCallback){
		TFoundInfo fi;
		fi.code = StopReq ? FI_CANCEL : FI_END;
		fi.filename = NULL;
		FoundCallback(UserData, this, fi);
	}
}

static int sort_comp( tnstr **s1, tnstr **s2 )
{
	return _tcsicmp((*s1)->c_str(), (*s2)->c_str());
}

//TODO: to common
void sort_filename(tnstr_vec &vec)
{
	return vec.sort(sort_comp);
}

bool TFindFileTask::FindDirs(const tchar *dir)
{
	bool dir_lastslash = IsLastSlash(dir);
	WIN32_FIND_DATA ffd;
	tnstrbuf path(dir);
	if (!dir_lastslash)
		path += _t(DIR_C);
	path += _t("*.*");
	HANDLE h = FindFirstFile(path, &ffd);
	if (h==INVALID_HANDLE_VALUE)
		return false;
	tnstr_vec dirs;
	tnstr_vec files;
	map< tnstr, DWORD > file_ts;
	do {
		if (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY){
			if (ffd.cFileName[0]=='.'){
				if (!ffd.cFileName[1] || (ffd.cFileName[1]=='.' && !ffd.cFileName[2])){
					continue;
				}
			}
			dirs.push(ffd.cFileName);
		} else {
			// found a file.
			if (!Masks.size() || CompareMask(Masks, ffd.cFileName)){
				if (Option & SRCH_SORT_NAME){
					files.push(ffd.cFileName);
					file_ts[ffd.cFileName] = ffd.ftLastWriteTime.dwLowDateTime;
				} else {
					path = dir;
					if (!dir_lastslash)
						path += _t(DIR_C);
					path += ffd.cFileName;
					FindStrTask->ReqFile(path, ffd.ftLastWriteTime.dwLowDateTime);
				}
				//dbw("path=%ws", path.c_str());
			}
		}
	} while (!StopReq && FindNextFile(h, &ffd));
	FindClose(h);

	if (Option & SRCH_SORT_NAME){
		sort_filename(files);
		foreach_tnstr_vec(files, it){
			if (StopReq)
				break;
			path = dir;
			if (!dir_lastslash)
				path += _t(DIR_C);
			path += *it;
			FindStrTask->ReqFile(path, file_ts[*it]);
		}
	}
	
	if (StopReq)
		return false;
	
	if (Option & SRCH_SUBDIR){
		// sub directories
		int i = 0;
		if (Option & SRCH_SORT_NAME){
			sort_filename(dirs);
		}
		foreach_tnstr_vec(dirs, d){
			tnstrbuf path(dir);
			if (!dir_lastslash)
				path += _t(DIR_C);
			path += *d;
			if (!FindDirs(path)){
				return false;
			}
			i++;
		}
	}
	return true;
}

#define	USE_STRMATCH	1
#define	USE_PATHMATCH	0

bool CompareMask(tnstr_vec &masks, const tchar *filename)
{
#if !USE_STRMATCH && !USE_PATHMATCH
	const tchar *ext = GetFileExtension(filename);
	if (!ext)
		return false;
#endif
	foreach_tnstr_vec(masks, mask){
#if USE_STRMATCH
		if (strmatch(filename, mask->c_str()))
			return true;
#elif USE_PATHMATCH
		if (PathMatchSpec(filename, mask->c_str()))
			return true;
#else
		// gq̂
		const tchar *mask_ext = GetFileExtension(*mask);
		if (_tcsicmp(mask_ext, ext)==0){
			return true;
		}
#endif
	}
	return false;
}

