//---------------------------------------------------------------------------

#include "tnlib.h"
#pragma hdrstop

#include "FindStrTask.h"
#include "SrchDef.h"
#include "FileCache.h"
#include "mmf.h"
#include "jtype.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)

static TFileCache FCache;

void TFindStrTask::Setup(const tchar *pattern, int option, FNFoundCallback callback, void *userdata)
{
	if (option&SRCH_WORD && !IsWordChar(Pattern)){
		option &= ~SRCH_WORD;	// reset word option if not word char included.
	}

	Lock();
	Pattern = pattern;
	nPattern = FCache.NormalizePattern(Pattern);
	nPatternLen = nPattern.length();
	Option = option;
	FoundCallback = callback;
	UserData = userdata;
	if (fs) delete fs;
	fs = new _tFindStr(Pattern, false);
	fsi = new _tFindStr(Pattern, true);
	Unlock();

	if (SubTask)
		SubTask->Setup(pattern, option&~SRCH_USE_CACHE, callback, userdata);
}
bool TFindStrTask::Start()
{
	if (!super::Start())
		return false;
	if (SubTask && !SubTask->Start()){
		Stop();
		return false;
	}
	return true;
}
void TFindStrTask::Stop()
{
	Lock();
	ReqFile(NULL, 0);
	FoundCallback = NULL;
	if (SubWaitEvent)
		WakeParentTask();
	Unlock();
	super::Stop();

	if (SubTask)
		SubTask->Stop();
}
void TFindStrTask::StopCallback()
{
	FoundCallback = NULL;

	if (SubTask)
		SubTask->StopCallback();
}
void TFindStrTask::Done()
{
	ReqFile(NULL, 0);

	if (SubTask)
		SubTask->Done();
}
void TFindStrTask::ReqFile(const tchar *filename, int timehash)
{
	Lock();
	while (!FileNames.push(filename)){
		Unlock();
		Sleep(1);
		Lock();
	}
	TimeHashes.push(timehash);
	sem.Release();
	Unlock();
}
void TFindStrTask::Execute()
{
	for (;!StopReq;){
		if (!sem.Wait()){
			break;	// error?
		}
		Lock();
		tnstr *filename = FileNames.pop();
		int th = TimeHashes.front(); TimeHashes.pop();
		Unlock();
		if (!filename || filename->empty()){
			// quit?
			break;
		}
		FindString(*filename, th);
		delete filename;
	}
	if (ParentTask){
		// I am sub task.
		
	} else {
		if (FoundCallback){
			TFoundInfo fi;
			fi.code = StopReq ? FI_CANCEL : FI_END;
			fi.filename = NULL;
			FoundCallback(UserData, this, fi);
		}
	}
}

void TFindStrTask::FindString(const tchar *filename, int timehash)
{
	if (!FoundCallback)
		return;

	int findret = 0;
	if (Option & SRCH_USE_CACHE){
		tnstr nfilename;
		findret = FCache._Find(filename, nfilename, timehash, nPattern, nPatternLen, Option&SRCH_USE_TS);
		if (findret==0){
			return;
		}
		if (findret==1){
			if (Option & SRCH_ONLY_FILENAME){
				TFoundInfo fi;
				fi.code = FI_FOUND;
				fi.filename = filename;
				fi.line = 0;
				fi.col = 0;
				fi.text = _t("");
				FoundCallback(UserData, this, fi);
				return;
			}
		}
		if (findret==-2){
			// file updated. erase cache
			FCache.__Erase(nfilename);
		}
		if (SubTask){
			//TODO: ꂾƈԍŌReqFile()̌ʂ擾łȂ̂ŒӁI
			if (SubWaiting){
				WaitSubTask();
				if (!FoundCallback){
					SubWaiting = false;
					return;
				}
			}
			SubTask->ReqFile(filename, timehash);

			// Update cache.
			if (svFindRet<0){
				if (!StopReq){
					tnstr nfilename;
					FCache._Add(svFileName, nfilename, svTimeHash, nPattern, nPatternLen, Option, SubTask->LastCaseIgnHit);
				}
			}

			svFindRet = findret;
			svTimeHash = timehash;
			svFileName = filename;
			SubWaiting = true;
			return;
		}
	}

	tnstr line;
	int len;
	int line_ct = 0;
	int found_ct = 0;
	bool cignhit = false;
	bool caseign = (Option & (SRCH_USE_CACHE|SRCH_CASE_IGNORE)) ? true : false;
	bool casecheck = (Option & SRCH_USE_CACHE) && !(Option & SRCH_CASE_IGNORE) ? true : false;

#if 0
	// Memory Mapped File access //
	for(;;){
		TMmfFile mmf;
		if (!mmf.open(filename, true)){
			break;
		}
		unsigned char *file = (unsigned char*)mmf.map(0, 0, true);
		if (!file)
			break;
		int filesize = mmf.filesize();

		// char code //
		unsigned short bom = 0;
		if (filesize>=2){
			unsigned short us = *(unsigned short*)file;
			switch (us){
				case 0xFEFF:	// UTF-16(LE)
					bom = us;
					file += 2;
					filesize -= 2;
					break;
				case 0xFFFE:	// UTF-16(BE)
					bom = us;
					file += 2;
					filesize -= 2;
					break;
				case 0xBBEF:
					if (filesize>=3 && file[3]==0xBF){
						// UTF-8
						bom = us;
						file += 3;
						filesize -= 3;
					}
					break;
			}
		}
		
		
			
		mmf.close();
	}
#endif

	TIFile tif;
	if (tif.open(filename)){
		//TODO: open error
		goto jerr;
	}

	while ((len=tif.getline(line))>=0){
		if (StopReq)
			break;
		const tchar *top = line;
		const tchar *p;
jretry:
		if (caseign)
			p = fstristr(top, len, *fsi);
		else
			p = fstrstr(top, len, *fs);
		if (p){
			cignhit = true;
			if (casecheck){
				p = fstrstr(top, len, *fs);
				if (!p){
					goto jcont;
				}
			}
			if (Option & (SRCH_HEAD|SRCH_WORD)){
				if (p!=line.c_str()){
					if (IsWordChar(p[-1])){
						len -= STR_DIFF(p, top)+1;
						top = p + 1;
						goto jretry;
					}
				}
				if (Option & SRCH_WORD){
					if (IsWordChar(p[nPatternLen])){
						len -= STR_DIFF(p, top)+1;
						top = p + 1;
						goto jretry;
					}
				}
			}
			if (FoundCallback){
				TFoundInfo fi;
				fi.code = FI_FOUND;
				fi.filename = found_ct==0?filename:NULL;
				fi.line = line_ct;
				fi.col = STR_DIFF(p, line.c_str());
				fi.text = line;
				FoundCallback(UserData, this, fi);
				found_ct++;
			}
			if (Option & SRCH_ONLY_FILENAME){
				break;
			}
		}
	jcont:
		line_ct++;
	}
jerr:
	if (ParentTask){
		LastCaseIgnHit = cignhit;
		WakeParentTask();
	}
	if (findret<0){
		if (!StopReq){
			tnstr nfilename;
			FCache._Add(filename, nfilename, timehash, nPattern, nPatternLen, Option, cignhit);
		}
	}
}

bool FindStrSaveCache()
{
	return FCache.Save();
}
bool FindStrLoadCache()
{
	return FCache.Load();
}
bool FindStrIsCacheModified()
{
	return FCache.IsModified();
}


