//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop
#include "tnlib.h"
#include "vclcom.h"

#include "MainFrm.h"
#include "SrchDef.h"
#include "prgcommon.h"
#include "prgprofile.h"
#include "pfs.h"
#include "SettingDlg.h"
#include "wcommon.h"
#include "MainCmd.h"
#include "prgstrlib.h"
#include "prgconfig.h"
#include "UIMain.h"
#include "SquFrm.h"
#include "SquIF.h"
#include "SearchSetDlg.h"
#include "AboutDlg.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "TntMenus"
#pragma link "TntComCtrls"
#pragma link "TntExtCtrls"
#pragma link "TntStdCtrls"
#pragma resource "*.dfm"
TMooreaMainForm *MooreaMainForm;
HINSTANCE hInstance;

// Prototypes //
void LoadHistory(TTntComboBox *cb, const tchar *pfs);
void SaveHistory(TTntComboBox *cb, const tchar *pfs);
bool AddToHistory(TTntComboBox *cb, bool caseign);

//---------------------------------------------------------------------------
__fastcall TMooreaMainForm::TMooreaMainForm(TComponent* Owner)
	: TTntForm(Owner)
{
	UIMain = TUIMain::CreateInstance(this);
	squfrm = new TSquFrame(pnlSqu, UIMain);

	MsgPastTime = -1;
	MsgCount = -1;

	// View //
	AlwaysOnTop = false;

	UIChanging = false;

	CacheLoaded = false;
}
//---------------------------------------------------------------------------
__fastcall TMooreaMainForm::~TMooreaMainForm()
{
	__delete(squfrm);
}
void __fastcall TMooreaMainForm::FormShow(TObject *Sender)
{
	Initialize();

	LoadOptions();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::TntFormCloseQuery(TObject *Sender,
	  bool &CanClose)
{
	if (CanClose){
		if (CacheLoaded && cmdIsCacheModified()){
			int ret = MsgBox(_t("Save the cache data?"), Caption.c_bstr(), MB_YESNOCANCEL|MB_ICONQUESTION);
			if (ret==IDCANCEL){
				CanClose = false;
				return;
			}
			if (ret==IDYES){
				squfrm->StopSearchAll();
				cmdSaveCache();
			}
		}
	}
	SaveOptions();
	Cleanup();
}
//---------------------------------------------------------------------------
// Main Menu //
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miOpenEditorClick(TObject *Sender)
{
	KeyEnter();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miOpenToolClick(TObject *Sender)
{
	if (squi()->GetItemCount()>0){
		squi()->OpenTool();
	}
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miLoadCacheClick(TObject *Sender)
{
	if (squfrm->IsSearchingAll())
		return;
	SetMsg(NULL, _t("Loading..."));
	Update();
	cmdLoadCache();
	miLoadCache->Enabled = false;
	cbUseCache->Checked = true;
	SetMsg(NULL, NULL);
	CacheLoaded = true;
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miSaveCacheClick(TObject *Sender)
{
	if (squfrm->IsSearchingAll())
		return;
	cmdSaveCache();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miExitClick(TObject *Sender)
{
	Close();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miClearClick(TObject *Sender)
{
	ClearEdit();
	edPattern->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miToUpperClick(TObject *Sender)
{
	tnstr str = edPattern->Text.c_bstr();
	LowerToUpper( str.c_str() );
	edPattern->Text = str;
	EditChangeNotify();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miToLowerClick(TObject *Sender)
{
	tnstr str = edPattern->Text.c_bstr();
	CharLower( str.c_str() );
	edPattern->Text = str;
	EditChangeNotify();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miCopyResultClick(TObject *Sender)
{
	CopyResult();
}
//---------------------------------------------------------------------------
// Search menu
void __fastcall TMooreaMainForm::miSearchClick(TObject *Sender)
{
	miSearchClipboard->Enabled = IsClipboardText();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miSearchClipboardClick(TObject *Sender)
{
	SearchClipboard();
	edPattern->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miSearchSetClick(TObject *Sender)
{
	SearchSetConfig();
}
//---------------------------------------------------------------------------
// View menu
void __fastcall TMooreaMainForm::View1Click(TObject *Sender)
{
	miAlwaysOnTop->Checked = AlwaysOnTop;
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miAlwaysOnTopClick(TObject *Sender)
{
	SetAlwaysOnTop(!AlwaysOnTop);
}
//---------------------------------------------------------------------------
// Tools menu
void __fastcall TMooreaMainForm::miOptionsClick(TObject *Sender)
{
	DoSetting();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miSortByFilenameClick(TObject *Sender)
{
	ToggleMenuCheck(miSortByFilename);
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miOnlyFilenameClick(TObject *Sender)
{
	ToggleMenuCheck(miOnlyFilename);
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miCheckTimestampClick(TObject *Sender)
{
	ToggleMenuCheck(miCheckTimestamp);
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miSubDirectoriesClick(TObject *Sender)
{
	ToggleMenuCheck(miSubDirectories);
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miNewTabClick(TObject *Sender)
{
	squfrm->NewTab();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miCloseTabClick(TObject *Sender)
{
	if (squfrm->GetTabCount()<=1)
		Close();
	else
		squfrm->CloseTab();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::actToResultExecute(TObject *Sender)
{
	squi()->ResultSetFocus();
}
//---------------------------------------------------------------------------
// Component Event Handers //
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::edPatternKeyDown(TObject *Sender,
	  WORD &Key, TShiftState Shift)
{
	bool redirect = false;
#if 0
	if (Key==VK_ESCAPE){
		if (edPattern->DroppedDown){
			edPattern->DroppedDown = false;
			Key = 0;
		} else {
			ClearEdit();
			Key = 0;
		}
	} else
#endif
	if (Key==VK_DOWN){
#if 0
		if (Shift==sscNone){
			ResultSelDown();
			Key = 0;
		} else
#endif
		if (edPattern->DroppedDown || Shift==sscAlt){
			// for editor
		} else
		if (Shift==sscCtrl){
			// Ctrl+
			NextSelFile();
			Key = 0;
		} else {
			redirect = true;
		}
	} else
	if (Key==VK_UP){
		if (edPattern->DroppedDown){
			// for editor
		} else
		if (Shift==sscNone){
			ResultSelUp();
			Key = 0;
		} else
		if (Shift==sscCtrl){
			// Ctrl+
			PrevSelFile();
			Key = 0;
		} else {
			redirect = true;
		}
	} else
	if (Key==VK_HOME || Key==VK_END){
		if (Shift==sscNone || Shift==sscShift){
			// for editor
		} else {
			redirect = true;
		}
	} else
	if (Key==VK_NEXT || Key==VK_PRIOR){
		redirect = true;
	}
	if (redirect){
		SendMessage(squi()->GetResultHandle(), WM_KEYDOWN, Key, 0);
		Key = 0;
	}
}
void __fastcall TMooreaMainForm::edPatternChange(TObject *Sender)
{
	EditChangeNotify();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::edDirChange(TObject *Sender)
{
	DirChangeNotify();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::edMaskChange(TObject *Sender)
{
	MaskChangeNotify();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::cbCaseSenseClick(TObject *Sender)
{
	EditChangeNotify();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::cbHeadClick(TObject *Sender)
{
	if (cbHead->Checked)
		cbWord->Checked = false;
	EditChangeNotify();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::cbWordClick(TObject *Sender)
{
	if (cbWord->Checked)
		cbHead->Checked = false;
	EditChangeNotify();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::btnOptionClick(TObject *Sender)
{
	//
	POINT pt;
	pt.x = 0;
	pt.y = btnOption->Height;
	::ClientToScreen(btnOption->Handle, &pt);
	pmOption->Popup(pt.x, pt.y);
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::cmOpenEditorClick(TObject *Sender)
{
	KeyEnter();
}
//---------------------------------------------------------------------------
// Result List
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::lvResultDblClick(TObject *Sender)
{
	KeyEnter();
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::miAboutClick(TObject *Sender)
{
	AboutDialog = new TAboutDialog(this);
	AboutDialog->ShowModal();
	delete AboutDialog;
}
//---------------------------------------------------------------------------
// ActionList
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::actTabExecute(TObject *Sender)
{
	TAction *action = (TAction*)Sender;
	squfrm->SelectTab(action->Tag);
}
//---------------------------------------------------------------------------
// Misc events
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::tmPeriodicTimer(TObject *Sender)
{
	HWND hwnd = GetFocus();
	if (cbCaseSense->Handle==hwnd
		|| cbHead->Handle==hwnd
		|| cbWord->Handle==hwnd
		|| cbRegular->Handle==hwnd
		|| cbUseCache->Handle==hwnd
		){
		edPattern->SetFocus();
	}
}
//---------------------------------------------------------------------------
void __fastcall TMooreaMainForm::tmUpdateTimer(TObject *Sender)
{
	tmUpdate->Enabled = false;
	StartSearch();
}
//---------------------------------------------------------------------------
void TMooreaMainForm::Initialize()
{
	InitTNWinLib(hInstance);
	PrgInit(hInstance, Handle);
	InitProfile();

	Caption = _t("Moorea!");
	
	squfrm->Setup();
	squfrm->RestoreTabs();
	if (squfrm->GetTabCount()==0){
		squfrm->NewTab();
	}

	squfrm->RestoreActiveTab();

	LoadHistory(edPattern, PFS_PATTERN_HISTORY);
	LoadHistory(edDir, PFS_DIR_HISTORY);
	LoadHistory(edMask, PFS_MASK_HISTORY);

	Application->OnIdle = IdleHandler;

	TRegKey *reg = prof.OpenKey(PFS_MAIN);
	LoadForm(reg, this, (HWND)NULL);
	SetAlwaysOnTop(reg->ReadInteger(PFS_VIEWTOP, AlwaysOnTop));
	delete reg;
}
void TMooreaMainForm::Cleanup()
{
	TRegKey *reg = prof.OpenKey(PFS_MAIN);
	SaveForm(reg, this, (HWND)NULL);
	reg->WriteInteger(PFS_VIEWTOP, AlwaysOnTop);
	delete reg;

	ClearEdit();

	squfrm->Cleanup();

	__delete(UIMain);
}
void TMooreaMainForm::LoadOptions()
{
	UIChanging = true;
	TRegKey *reg = prof.OpenKey(PFS_MAIN);
	cbCaseSense->Checked = reg->ReadInteger(PFS_CASE, cbCaseSense->Checked);
	cbHead->Checked = reg->ReadInteger(PFS_HEAD, cbHead->Checked);
	cbWord->Checked = reg->ReadInteger(PFS_WORD, cbWord->Checked);
	delete reg;
	UIChanging = false;
}
void TMooreaMainForm::SaveOptions()
{
	TRegKey *reg = prof.OpenKey(PFS_MAIN);
	reg->WriteInteger(PFS_CASE, cbCaseSense->Checked);
	reg->WriteInteger(PFS_HEAD, cbHead->Checked);
	reg->WriteInteger(PFS_WORD, cbWord->Checked);
	delete reg;
}
void __fastcall TMooreaMainForm::IdleHandler(TObject *sender, bool &done)
{
	IdleProc(done);
}
//---------------------------------------------------------------------------
// Edit Control //
void TMooreaMainForm::ClearEdit()
{
	if (edPattern->Text.Length()>=MIN_SEARCH_LENGTH){
		if (AddToHistory(edPattern, false))
			SaveHistory(edPattern, PFS_PATTERN_HISTORY);
	}
	edPattern->Text = "";
	squi()->StopSearch();
}
//---------------------------------------------------------------------------
// View //
void TMooreaMainForm::SetAlwaysOnTop(bool enabled)
{
	if (AlwaysOnTop==enabled)
		return;

	AlwaysOnTop = enabled;

	SetWindowPos( Application->Handle, AlwaysOnTop ? HWND_TOPMOST : HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE );
	if (AlwaysOnTop)
		Application->NormalizeTopMosts();
}

//---------------------------------------------------------------------------
// Result List //
void TMooreaMainForm::ResultSelDown()
{
	squi()->ResultSelDown();
}
void TMooreaMainForm::ResultSelUp()
{
	squi()->ResultSelUp();
}
void TMooreaMainForm::NextSelFile()
{
	squi()->NextSelFile();
}
void TMooreaMainForm::PrevSelFile()
{
	squi()->PrevSelFile();
}
void TMooreaMainForm::CopyResult()
{
	squi()->CopyResult();
}
void TMooreaMainForm::IdleProc(bool &done)
{
	if (squfrm){
		done = squfrm->IdleProc();
	}
}
// Key command //
void TMooreaMainForm::KeyEnter()
{
	if (squi()->GetItemCount()>0){
		squi()->KeyEnter();
	} else {
		StartSearch();
	}
}
// Event //
void TMooreaMainForm::EditChangeNotify()
{
	if (!UIChanging){
		StartSearch();
	}
}
void TMooreaMainForm::DirChangeNotify()
{
	if (UIChanging)
		return;
	tmUpdate->Enabled = false;
	tmUpdate->Enabled = true;
}
void TMooreaMainForm::MaskChangeNotify()
{
	if (UIChanging)
		return;
	tmUpdate->Enabled = false;
	tmUpdate->Enabled = true;
}
void TMooreaMainForm::SearchClipboard()
{
	tnstr text = GetClipboardText(Handle);
	if (text.exist()){
		if (edPattern->Text!=text){
			edPattern->Text = text;
			edPattern->SelStart = 999999;
			edPattern->SelLength = 0;
			EditChangeNotify();
		}
	}
}
void TMooreaMainForm::SearchSetConfig()
{
	TSearchSetDialog *dlg = new TSearchSetDialog(this);
	dlg->ShowModal();
	delete dlg;
}
void TMooreaMainForm::DoSetting()
{
	TSettingDialog *dlg = new TSettingDialog(this);
	dlg->ShowModal();
	delete dlg;
}

TSquIF *TMooreaMainForm::squi()
{
	return squfrm->GetActive();
}
#if 0
TSearchProcI *TMooreaMainForm::srchProc()
{
	return squi()->GetSrchProc();
}
#endif

//-------------------------------------------------------------------------
// message
//-------------------------------------------------------------------------
void TMooreaMainForm::ClearMsg()
{
	lbStatus->Caption = "";
//	if (!lbWarning->Caption.IsEmpty())
//		lbWarning->Visible = true;
}
#if 0
void TMooreaMainForm::SetMsg( int msgno, int pasttime, int count )
{
//	SetMsg(GetStateMessage( msgno ), pasttime, count);
}
#endif
// count:
//	-1  - invalid
//	>=0 - found count
//  <-1 - over count than -count.
void TMooreaMainForm::SetMsg(class TSquIF *sender, const tchar *msg, int pasttime, int count)
{
	if (sender && sender!=squi())
		return;

	if (msg){
		if (msg[0])
			lbWarning->Visible = false;
		else
			if (!lbWarning->Caption.IsEmpty())
				lbWarning->Visible = true;
		lbStatus->Caption = msg;
	} else {
		lbStatus->Caption = "";
		if (!lbWarning->Caption.IsEmpty())
			lbWarning->Visible = true;
	}

	if (MsgCount!=count || MsgPastTime!=pasttime){
		MsgCount = count;
		MsgPastTime = pasttime;
		updateMessageR();
	}
}

void TMooreaMainForm::SetWarning(class TSquIF *sender, const tchar *msg)
{
	if (sender && sender!=squi())
		return;

	if (msg && msg[0]){
		lbWarning->Caption = msg;
		if (lbStatus->Caption.IsEmpty())
			lbWarning->Visible = true;
	} else {
		lbWarning->Caption = "";
		lbWarning->Visible = false;
	}
}

void TMooreaMainForm::SetFileInfo(class TSquIF *sender, const tchar *str)
{
	if (sender && sender!=squi())
		return;

	if (str && str[0]){
		lbFileInfo->Caption = str;
	} else {
		lbFileInfo->Caption = _t("");
	}
}


void TMooreaMainForm::updateMessageR()
{
	tnstr text;
	if (MsgCount!=-1){
		text += _t("[");
		if (MsgCount>=0){
			text += itocs(MsgCount);
		} else {
			text += _t(">");
			text += itocs(-MsgCount);
		}
		text += _t("]");
	}
	if (MsgPastTime>=0){
		text += tnsprintf(_t("[%1d.%02d]"), MsgPastTime/1000, (MsgPastTime%1000)/10);
	}
	lbTime->Caption = text;
}

// Search Control //
void TMooreaMainForm::StartSearch()
{
	int option = 0;
	if (!cbCaseSense->Checked)
		option |= SRCH_CASE_IGNORE;
	if (cbHead->Checked)
		option |= SRCH_HEAD;
	if (cbWord->Checked)
		option |= SRCH_WORD;
	if (cbRegular->Checked)
		option |= SRCH_REGULAR;
	if (miSortByFilename->Checked)
		option |= SRCH_SORT_NAME;
	if (miOnlyFilename->Checked)
		option |= SRCH_ONLY_FILENAME;
	if (cbUseCache->Checked){
		option |= SRCH_USE_CACHE;
		miLoadCache->Enabled = false;
	}
	if (miCheckTimestamp->Checked)
		option |= SRCH_USE_TS;
	if (miSubDirectories->Checked)
		option |= SRCH_SUBDIR;
	tnstr pat(edPattern->Text);
	if (squi()->StartSearch(pat, edDir->Text, edMask->Text, option)){
		SaveHistories();
	}
}
void TMooreaMainForm::SaveHistories()
{
	if (AddToHistory(edDir, true)
		|| AddToHistory(edMask, false)
		){
		SaveHistory(edDir, PFS_DIR_HISTORY);
		SaveHistory(edMask, PFS_MASK_HISTORY);
	}
}
//---------------------------------------------------------------------------
// Notify
//---------------------------------------------------------------------------
void TMooreaMainForm::NotifyTabChanged()
{
	squi()->UpdateMsg();

	TSearchParams p;
	squi()->GetSearchParams(p);
	if (p.pat.exist()){
		UIChanging = true;
		SetSearchParams(p);
		UIChanging = false;
	}
}

void TMooreaMainForm::SetSearchParams(const TSearchParams &p)
{
	edPattern->Text = p.pat;
	edDir->Text = p.dir;
	edMask->Text = p.mask;
	cbCaseSense->Checked = p.option & SRCH_CASE_IGNORE ? false : true;
	miOnlyFilename->Checked = p.option & SRCH_ONLY_FILENAME ? true : false;
}

//---------------------------------------------------------------------------

// Utility //
void LoadHistory(TTntComboBox *cb, const tchar *pfs)
{
	cb->Items->Clear();

	tnstr_vec strs;
	ReadHistory(pfs, strs);
	cb->Items->BeginUpdate();
	foreach_tnstr_vec(strs, s){
		cb->Items->Add(*s);
	}
	cb->Items->EndUpdate();
	if (strs.size()>0){
		cb->Text = strs[0];
	}
}
void SaveHistory(TTntComboBox *cb, const tchar *pfs)
{
	tnstr_vec strs;
	for (int i=0;i<cb->Items->Count;i++){
		strs.push(cb->Items->Strings[i].c_bstr());
	}
	WriteHistory(pfs, strs);
}

int IndexOf(TTntComboBox *cb, const tchar *text, bool caseign);
int Find(TTntComboBox *cb, const tchar *text, bool caseign);

bool AddToHistory(TTntComboBox *cb, bool caseign)
{
	if (cb->Focused())
		return false;

	if (cb->Text.IsEmpty())
		return false;
	int index = Find(cb, cb->Text, caseign);
	if (index==0){
		return false;	// no change
	}
	if (index<0){
		// new item
		cb->Items->Insert(0, cb->Text);
	} else {
		cb->Items->Exchange(0, index);
		cb->Text = cb->Items->Strings[0];	// VCL bug?
	}
	return true;	// updated.
}

//TODO: ėp֐
#include "ffstr.h"
int IndexOf(TTntComboBox *cb, const tchar *text, bool caseign)
{
	int index;
	_tFindStr fs(cb->Text, caseign);
	for (index=0;index<cb->Items->Count;index++){
		const WideString &ws = cb->Items->Strings[index];
		if (caseign){
			if (fstristr(ws, ws.Length(), fs))
				return index;
		} else {
			if (fstrstr(ws, ws.Length(), fs))
				return index;
		}
	}
	return -1;
}

int Find(TTntComboBox *cb, const tchar *text, bool caseign)
{
	int index;
	for (index=0;index<cb->Items->Count;index++){
		const WideString &ws = cb->Items->Strings[index];
		if (caseign){
			if (!_tcsicmp(ws.c_bstr(), cb->Text.c_bstr()))
				return index;
		} else {
			if (!_tcscmp(ws.c_bstr(), cb->Text.c_bstr()))
				return index;
		}
	}
	return -1;
}





