//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop
#include "tnlib.h"

#include "vclcom.h"
#include "SearchSetDlg.h"
#include "SearchSetData.h"
#include "SearchSetEditDlg.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "TntMenus"
#pragma link "TntStdCtrls"
#pragma resource "*.dfm"

#define	INX_SHORTCUT	0
#define	INX_MASK		1
#define	INX_DIR			2

void _ParseJumpName(const tchar *text, tnstr &name, tnstr &shortcut);
tnstr _BuildJumpName(const tchar *name, const tchar *shortcut);

TSearchSetDialog *SearchSetDialog;
//---------------------------------------------------------------------------
__fastcall TSearchSetDialog::TSearchSetDialog(TComponent* Owner)
	: TTntForm(Owner)
{
	Shortcuts = new TTntStringList;
}
//---------------------------------------------------------------------------
_fastcall TSearchSetDialog::~TSearchSetDialog()
{
	if (Shortcuts)
		delete Shortcuts;
}
void __fastcall TSearchSetDialog::TntFormShow(TObject *Sender)
{
  if ( !JumpList->Selected ){
	if (JumpList->Items->Count>0){
		JumpList->Selected = JumpList->Items->Item[0];
		JumpList->ItemFocused = JumpList->Selected;
	}
  }
  VCL_CENTER();
}
//---------------------------------------------------------------------------
void __fastcall TSearchSetDialog::miUpClick(TObject *Sender)
{
	MoveUp();
}
//---------------------------------------------------------------------------
void __fastcall TSearchSetDialog::miDownClick(TObject *Sender)
{
	MoveDown();
}
//---------------------------------------------------------------------------
void __fastcall TSearchSetDialog::EditPathBtnClick(TObject *Sender)
{
	TTntListItem *item = JumpList->Selected;
	if (!item)
		return;
	int index = item->Index;

	SearchSetEditDialog = new TSearchSetEditDialog( this );
	SearchSetEditDialog->Setup(item->Caption, item->SubItems->Strings[INX_DIR], item->SubItems->Strings[INX_MASK], item->SubItems->Strings[INX_SHORTCUT]);
	if ( SearchSetEditDialog->ShowModal() == mrOk ){
#if 0	//TODO:
		if ( SearchSetEditDialog->Path->Text.Length() == 0 ){
			item->Delete();
		} else {
			item->Caption = SearchSetEditDialog->Title->Text;
			item->SubItems->Strings[0] = SearchSetEditDialog->GetShortcut();
			item->SubItems->Strings[1] = SearchSetEditDialog->Path->Text;
			CheckShortcutDup(item->Index);
		}

		if (!JumpList->Selected){
			if (index>=JumpList->Items->Count)
				index = JumpList->Items->Count-1;
			if (index>=0)
				JumpList->Selected = JumpList->Items->Item[index];
		}
#endif
	}
	delete SearchSetEditDialog;
}
//---------------------------------------------------------------------------
void __fastcall TSearchSetDialog::RemoveBtnClick(TObject *Sender)
{
	if ( JumpList->Selected )
	{
		int selindex = JumpList->Selected->Index;	// save current index

		JumpList->Selected->Delete();	// remove

		if (selindex>=JumpList->Items->Count)
			selindex = JumpList->Items->Count-1;
		if (selindex>=0){	// select item if current item is avaliable
			JumpList->Selected = JumpList->Items->Item[selindex];
		}
	}
}
//---------------------------------------------------------------------------
void __fastcall TSearchSetDialog::btnUpClick(TObject *Sender)
{
	MoveUp();
}
//---------------------------------------------------------------------------
void __fastcall TSearchSetDialog::btnDownClick(TObject *Sender)
{
	MoveDown();
}
//---------------------------------------------------------------------------
void TSearchSetDialog::AddItems(TTntStrings *items)
{
	for (int i=0;i<items->Count;i++){
		AddItem(items->Values[items->Names[i]].c_bstr(), items->Names[i].c_bstr());
	}
}
void TSearchSetDialog::AddItem(const tchar *name, const tchar *filename)
{
	TTntListItem *item = JumpList->Items->Add();
	tnstr _name;
	tnstr shortcut;
	_ParseJumpName(name, _name, shortcut);
	item->Caption = _name;
	item->SubItems->Add(shortcut);
	item->SubItems->Add(filename);
}
// If find the same shortcut of index 'pri', then disable the found shortcut.
bool TSearchSetDialog::CheckShortcutDup(int pri)
{
	tnstr pri_shortcut = JumpList->Items->Item[pri]->SubItems->Strings[INX_SHORTCUT].c_bstr();
	for (int i=0;i<JumpList->Items->Count;i++){
		if (i==pri)
			continue;
		TTntListItem *li = JumpList->Items->Item[i];
		if (pri_shortcut == li->SubItems->Strings[INX_SHORTCUT]){
			li->SubItems->Strings[INX_SHORTCUT] = "";
			return true;	// found
		}
	}
	return false;	// no duplicated shortcut
}
void TSearchSetDialog::GetItems(TTntStrings *items)
{
	//TODO: dir=Ăꍇ삵Ȃ
	for (int i=0;i<JumpList->Items->Count;i++){
		TTntListItem *li = JumpList->Items->Item[i];
		tnstr s(li->SubItems->Strings[1]);
		s += _t("=");
		s += _BuildJumpName(li->Caption.c_bstr(), li->SubItems->Strings[0].c_bstr());
		items->Add(s);
	}
}
int TSearchSetDialog::GetSelIndex()
{
	if (JumpList->Selected)
		return JumpList->Selected->Index;
	return -1;
}
void _ParseJumpName(const tchar *text, tnstr &name, tnstr &shortcut)
{
	ParseJumpName(text, name, shortcut);
	if (shortcut.Length()>1 && shortcut[1]=='^'){
		shortcut = tnstr(_t("Ctrl+")) + &shortcut[2];
	}
}
tnstr _BuildJumpName(const tchar *name, const tchar *shortcut)
{
	tnstr _shortcut(shortcut);
	if (!_tcsncmp(shortcut, _t("Ctrl+"), 5)){
		_shortcut = tnstr(_t("^"))+&shortcut[5];
	}
	return BuildJumpName(name, _shortcut.c_str());
}
//---------------------------------------------------------------------------
void TSearchSetDialog::MoveUp()
{
	int i = JumpList->Selected->Index;
	if ( (i == 0) || (JumpList->Items->Count < 2) ) return;
	ListItemExchange(JumpList, i, i-1);
	JumpList->Selected = JumpList->Items->Item[i-1];
	JumpList->Selected->MakeVisible(false);
}
void TSearchSetDialog::MoveDown()
{
	int i = JumpList->Selected->Index;
	if ( (JumpList->Items->Count<2) || (i>=JumpList->Items->Count-1) ) return;
	ListItemExchange(JumpList, i, i+1 );
	JumpList->Selected = JumpList->Items->Item[i+1];
}

