//---------------------------------------------------------------------------

#include "tnlib.h"
#pragma hdrstop

#include "FindFileTask.h"
#include "SrchProc.h"
#include "jtype.h"
//#include "strlib_sub.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)

#define	MIN_UPDATE_INTERVAL		330		// [mSec] \XVŏԊu(Ō߂邱Ƃł͂ȂCEEEj

//#define	MAX_FILENAME_POOL	256

// thread safe string queue

#if 0
enum {
	ST_SEARCHING = 0x0001,
};
#endif

struct TFileLocItem {
	int line;
	int col;
	tnstr text;
	TFileLocItem(int _line, int _col, const tchar *_text)
		:line(_line)
		,col(_col)
		,text(_text)
	{
	}
};

struct TFileItem {
	tnstr filename;
	int status;	// ST_SEARCHING
	int numlines;
	FlexObjectArray<TFileLocItem> lines;
	TFileItem(const tchar *_filename, int _status=ST_SEARCHING)
		:filename(_filename)
		,status(_status)
	{
		numlines = -1;
	}
};

class TFileItemPool : public TMutex {
protected:
	//tnstr_que Files;
	FlexObjectArray<TFileItem> Files;
public:
	TFileItemPool()
		//:Files(MAX_FILENAME_POOL)
	{
	}
	~TFileItemPool()
	{
	}
	void lock(){ Lock(); }
	void unlock(){ Unlock(); }
	// Add a new file.
	void push(const tchar *s)
	{
		push(new TFileItem(s));
	}
	void push(TFileItem *s)
	{
		lock();
		Files.push(s);
		unlock();
	}
	// Add a new line.
	void push(int line, int loc, const tchar *text)
	{
		__assert(Files.size()>0);
		TFileLocItem *locitem = new TFileLocItem(line, loc, text);
		lock();
		TFileItem &item = Files[size()-1];
		item.lines.push(locitem);
		unlock();
	}
	bool get_nextline(int &fileinx, int &lineinx)
	{
		for (;;){
			if (fileinx>=Files.size())
				return false;	// no more data

			if (lineinx>=Files[fileinx].lines.size()){
				// end of the file
				if (fileinx==Files.size()-1){
					return false;	// no more data
				}
				lineinx = 0;
				fileinx++;
				continue;
			}
			return true;
		}
	}
	int abs_to_index(int abs_index, int &sub_index)
	{
		lock();
		for (int i=0;i<Files.size();i++){
			if (abs_index<=Files[i].lines.size()){
				// found!!
				sub_index = abs_index;
				unlock();
				return i;
			}
			abs_index -= Files[i].lines.size()+1;
		}
		unlock();
		return -1;
	}
	void clear()
	{
		Files.clear();
	}
#if 0
	TFileItem *pop()
	{
		lock();
		TFileItem *s = Files.pop();
		unlock();
		return s;
	}
#endif
	int size() const
		{ return Files.size(); }
	TFileItem &operator[] (int index)
		{ return Files[index]; }

	// Additional //
	int next_file_index(int start_abs_index)
	{
		lock();
		int sub_index;
		int index = abs_to_index(start_abs_index, sub_index);
		if (index<0 || index+1>=size()){
			unlock();
			return -1;
		}
		int ret = start_abs_index - sub_index + Files[index].lines.size() + 1;
		unlock();
		return ret;
	}
	int prev_file_index(int start_abs_index)
	{
		lock();
		int sub_index;
		int index = abs_to_index(start_abs_index, sub_index);
		if (index<0 || index==0){
			unlock();
			return -1;
		}
		int ret;
		if (sub_index==0)
			ret = start_abs_index - Files[index-1].lines.size() - 1;
		else
			ret = start_abs_index - sub_index;
		unlock();
		return ret;
	}
};

class TSearchProc {
public:
	TSearchViewI *View;
	TFileItemPool Pool;
	bool End;
	bool End2;
	bool Updated;
	int NextFileIndex;
	int NextLineIndex;
	DWORD StartTime;
	DWORD EndTime;
	DWORD PrevUpdate;
	tnstr Pattern;
public:
	TSearchProc(TSearchViewI *view)
	{
		View = view;
		End = false;
		End2 = false;
		Updated = false;
		NextFileIndex = 0;
		NextLineIndex = 0;
		PrevUpdate = 0;
	}
	bool Start(const tchar *pattern, const tchar *dir, const tchar *masks, int option);
	bool UpdateView();
	bool _UpdateView();
	bool GetFile(int abs_index, tnstr &filename, int &line, int &col)
	{
		int sub_index;
		int index = Pool.abs_to_index(abs_index, sub_index);
		if (index<0)
			return false;
		filename = Pool[index].filename;
		if (sub_index==0){
			line = 0;
			col = 0;
		} else {
			sub_index--;
			line = Pool[index].lines[sub_index].line;
			col = Pool[index].lines[sub_index].col;
		}
		return true;
	}
	bool GetFile(int abs_index, tnstr &filename)
	{
		int sub_index;
		int index = Pool.abs_to_index(abs_index, sub_index);
		if (index<0)
			return false;
		filename = Pool[index].filename;
		return true;
	}
	int GetSearchTime()
		{ return EndTime - StartTime; }
	int GetNextFileIndex(int start_abs_index)
	{
		return Pool.next_file_index(start_abs_index);
	}
	int GetPrevFileIndex(int start_abs_index)
	{
		return Pool.prev_file_index(start_abs_index);
	}
	const tchar *GetPattern() const
		{ return Pattern; }
protected:
	METHOD_FOUNDCALLBACK(FoundProc);
	void BeginUpdate();
	void EndUpdate();
};

static TFindFileTask *FFTask = NULL;
//static TSearchProc SearchProc;

bool TSearchProc::Start(const tchar *pattern, const tchar *dir, const tchar *masks, int option)
{
	if (!pattern || !pattern[0] || !pattern[1]){
		return false;
	}
	if (FFTask){
		FFTask->Stop();
	}
	End = false;
	End2 = false;
	Updated = false;
	NextFileIndex = 0;
	NextLineIndex = 0;
	Pool.clear();
	FFTask = new TFindFileTask();
	FFTask->Setup(pattern, dir, masks, option, FoundProc, this);
	if (!FFTask->Start()){
		//TODO: fatal error?
		return false;
	}
	EndTime = StartTime = GetTickCount();
	Pattern = pattern;
	return true;
}

bool TSearchProc::UpdateView()
{
	DWORD tick = GetTickCount();
	if (PrevUpdate!=0 && tick-PrevUpdate<MIN_UPDATE_INTERVAL){
		return false;
	}
	PrevUpdate = tick;
	return _UpdateView();
}

bool TSearchProc::_UpdateView()
{
	const int max_add_once = 1000;
	for (int i=0;;i++){
		if (!Pool.get_nextline(NextFileIndex, NextLineIndex)){
			Updated = false;
			EndUpdate();
			return true;	// all done
		}

		if (i==0){
			BeginUpdate();
		} else
		if (i>=max_add_once){
			EndUpdate();
			return false;	// not done
		}

		// Update the file.
		Pool.Lock();
		TFileItem &item = Pool[NextFileIndex];
		TFileLocItem &loc = item.lines[NextLineIndex];
		Pool.Unlock();

		if (NextLineIndex==0){
			View->AddFile(item.filename);
		}
		View->AddFileLoc(loc.line, loc.col, loc.text);

		NextLineIndex++;
	}
}

BEGIN_FOUNDCALLBACK(TSearchProc, FoundProc)
{
	switch (info.code){
		case FI_FOUND:
			if (info.filename){
				Pool.push(info.filename);
			}
			Pool.push(info.line, info.col, info.text);
			Updated = true;
			View->UpdateNotify();
			break;
		case FI_END:
		case FI_CANCEL:
			End = true;
			FFTask = NULL;
			EndTime = GetTickCount();
			View->EndNotify();
			break;
	}
	return 0;
}

void TSearchProc::BeginUpdate()
{
	View->BeginUpdate();
}
void TSearchProc::EndUpdate()
{
	View->EndUpdate();
}

// Interfaces //

TSearchProcI::TSearchProcI(TSearchViewI *view)
	:View(view)
{
	Proc = new TSearchProc(view);
}
TSearchProcI::~TSearchProcI()
{
	if (Proc)
		delete Proc;
}

bool TSearchProcI::StartSearch(const tchar *pattern, const tchar *dir, const tchar *masks, int option)
{
	if (Proc->Start(pattern, dir, masks, option)){
		View->UpdateStatus(ST_SEARCHING);
		View->UpdateTime(-1);
		return true;
	}
	return false;
}

void TSearchProcI::StopSearch()
{
	if (FFTask){
		FFTask->Stop();
		FFTask = NULL;
	}
	View->UpdateStatus(ST_NONE);
}

bool TSearchProcI::IdleProc()
{
	if (Proc->Updated){
		if (!Proc->UpdateView())
			return false;
	}
	if (Proc->End){
		if (!Proc->End2){
			View->UpdateStatus(ST_NONE);
			View->UpdateTime(Proc->GetSearchTime());
			Proc->End2 = true;
		}
	}
	return true;
}

bool TSearchProcI::GetFile(int index, tnstr &filename, int &line, int &col)
{
	return Proc->GetFile(index, filename, line, col);
}
bool TSearchProcI::GetFile(int index, tnstr &filename)
{
	return Proc->GetFile(index, filename);
}
int TSearchProcI::NextFileIndex(int start_abs_index)
{
	return Proc->GetNextFileIndex(start_abs_index);
}
int TSearchProcI::PrevFileIndex(int start_abs_index)
{
	return Proc->GetPrevFileIndex(start_abs_index);
}
const tchar *TSearchProcI::GetPattern() const
{
	return Proc->GetPattern();
}

int TSearchCommon::Search(const tchar *str)
{
	const tchar *top = str;
	const tchar *p;
jretry:
	if (option & SRCH_CASE_IGNORE)
		p = fstristr(top, _tcslen(top), fs);
	else
		p = fstrstr(top, _tcslen(top), fs);
	if (p){
		if (option & (SRCH_HEAD|SRCH_WORD)){
			if (p!=str){
				if (IsWordChar(p[-1])){
					top = p + 1;
					goto jretry;
				}
			}
			if (option & SRCH_WORD){
				if (IsWordChar(p[fs.n2])){
					top = p + 1;
					goto jretry;
				}
			}
		}
		pos = STR_DIFF(p, str);
		return 1;
	}
	return 0;
}

