//---------------------------------------------------------------------------

#ifndef SrchProcH
#define SrchProcH
//---------------------------------------------------------------------------

#include "SrchDef.h"

enum eSearchStatus {
	ST_NONE = 0,
	ST_SEARCHING = 0x0001,
};
class TSearchViewI {
public:
	int Status;
public:
	TSearchViewI()
	{
		Status = ST_NONE;
	}
	virtual void BeginUpdate() = 0;
	virtual void EndUpdate() = 0;
	virtual void AddFile(const tchar *filename) = 0;
	virtual void AddFileLoc(int line, int col, const tchar *text) = 0;
	virtual void UpdateTime(int msec) = 0;
	virtual void UpdateStatus(int status)
	{
		Status = status;
	}
	virtual void UpdateNotify() = 0;	// Note: called by another thread
	virtual void EndNotify() = 0;		// Note: called by another thread
	bool GetFile(int index, tnstr &filename, int &line, int &col);
	bool GetFile(int index, tnstr &filename);
};

class TSearchProcI {
protected:
	class TSearchViewI *View;	// referred
	class TSearchProc *Proc;	// owned
public:
	TSearchProcI(TSearchViewI *view);
	~TSearchProcI();
	bool StartSearch(const tchar *pattern, const tchar *dir, const tchar *masks, int option);
	void StopSearch();
	bool IdleProc();
	bool GetFile(int index, tnstr &filename, int &line, int &col);
	bool GetFile(int index, tnstr &filename);
	int NextFileIndex(int start_abs_index);
	int PrevFileIndex(int start_abs_index);
	const tchar *GetPattern() const;
};

#include "ffstr.h"

class TSearchCommon {
protected:
	int option;
	_tFindStr fs;
public:
	TSearchCommon(const tchar *pat, int _option)
		:option(_option)
		,fs(pat, _option & SRCH_CASE_IGNORE ? true : false)
	{
	}
	int Search(const tchar *str);

	int pos;
};

#endif

