//---------------------------------------------------------------------------

#include "tnlib.h"
#pragma hdrstop

#include <sys/stat.h>
#include "FileCache.h"
#include "filestr.h"
#include "wcommon.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)

#define	_NormalizePattern(buf, arg) \
	tnstr buf(arg); \
	CharLower(buf.c_str())

#define	_NormalizeFileName(buf) \
	CharLower(buf)

#define	FileInfoMap_insert(nfilename, tmhash, dirty) \
	FileInfoMap.insert(std::pair< tnstr, TCFileInfo >(nfilename, TCFileInfo(tmhash, dirty)))

#define	curtime()	CacheTime

TFileCache::TFileCache()
{
	LastPatInfoId = 0;
	CacheTime = 0;
	Modified = false;
	Saving = false;
}
TFileCache::~TFileCache()
{
}

void TFileCache::Clear()
{
	lock();
	CacheTime = 0;
	PatInfoMap.clear();
	PatList.clear();
	PatMap.clear();
	FileInfoMap.clear();
	unlock();
}

tnstr TFileCache::NormalizePattern(const tchar *str)
{
	_NormalizePattern(npat, str);
	return npat;
}

void TFileCache::Add(const tchar *filename, tmhash_t timehash, const tchar *srchpat, srchopt_t option, bool hit)
{
	_NormalizePattern(npat, srchpat);
	tnstr nfilename;
	_Add(filename, nfilename, timehash, npat, npat.length(), option, hit);
}

// nsrchpat must be normalized.
void TFileCache::_Add(const tchar *filename, tnstr &nfilename, tmhash_t timehash, const tchar *nsrchpat, int nsrchpatlen, srchopt_t option, bool hit)
{
	lock();

	AdvTime();

	TPatInfo *info = Find(filename, nfilename, nsrchpat, nsrchpatlen, 0);
	if (info){
		// found!!
		// timestamp̂ݍXV
		info->lasttime = curtime();
		FileInfoMap[nfilename] = TCFileInfo(timehash, true);
	} else {

		// remove the longer pattern.
		//TODO: "abcd"Ōς݂łƂA"abc"̌ʂŏ㏑KvL
		// GCł邩H

		// new pattern //
		patid_t patid;
		patmap_t::iterator it = PatMap.find(nsrchpat);
		if (it!=PatMap.end()){
			// already exist
			patid = it->second;
		} else {
			// new string
			patid = ++LastPatInfoId;
			PatList[patid] = nsrchpat;
			PatMap[nsrchpat] = patid;
		}
		PatInfoMap[nfilename].push_back(TPatInfo(patid, (hit?SRCH_OPT_HIT:0), curtime()));
		//dbw("add:%ws id=%d hit=%d", nfilename.c_str(), patid, hit);
		fileinfomap_t::iterator it2 = FileInfoMap.find(nfilename);
		if (it2!=FileInfoMap.end()){
			// already exist
			it2->second.tmhash = timehash;
			it2->second.dirty = true;
		} else {
			FileInfoMap_insert(nfilename, timehash, true);
		}
	}
	Modified = true;
	unlock();
}

void TFileCache::Erase(const tchar *filename)
{
	tnstr nfilename(filename);
	_NormalizeFileName(nfilename.c_str());
	__Erase(nfilename);
}

void TFileCache::__Erase(const tchar *nfilename)
{
	lock();
	PatInfoMap.erase(nfilename);
	FileInfoMap.erase(nfilename);
	unlock();
}

// return value:
//	1 : hit
//  0 : non-hit
//  -1: unknown()
//  -2: timestamp updated. only if option has SRCH_USE_TS
int TFileCache::Find(const tchar *filename, tmhash_t timehash, const tchar *srchpat, srchopt_t option)
{
	tnstr npat(srchpat);
	CharLower(npat.c_str());
	tnstr nfilename;
	return _Find(filename, nfilename, timehash, npat, npat.length(), option);
}

int TFileCache::_Find(const tchar *filename, tnstr &nfilename, tmhash_t timehash, const tchar *nsrchpat, int nsrchpatlen, srchopt_t option)
{
	TPatInfo *info = Find(filename, nfilename, nsrchpat, nsrchpatlen, option);
	if (info){
		if (option & SRCH_USE_TS){
			fileinfomap_t::iterator it = FileInfoMap.find(nfilename);
			if (it!=FileInfoMap.end()){
				if (it->second.tmhash!=timehash){
					return -2;	// file updated
				}
			}
		}
		return info->option & SRCH_OPT_HIT ? 1 : 0;
	}
	return -1;
}

TPatInfo *TFileCache::Find(const tchar *filename, tnstr &nfilename, const tchar *npat, int npatlen, srchopt_t option)
{
	// normalize filename
	nfilename = filename;
	_NormalizeFileName(nfilename.c_str());

	return __Find(nfilename, npat, npatlen, option);
}

//Note:
// you must lock until the return value unused.
// Not filename (nfilename).
TPatInfo *TFileCache::__Find(const tchar *nfilename, const tchar *npat, int npatlen, srchopt_t option)
{
	// find the pattern

	// PĊYp^[邩ׂ
	tnstr npatbuf;
	int orgnpatlen = npatlen;

	TAutoLock lock(Mutex);
	
	while (npatlen>0){
		patmap_t::iterator it = PatMap.find(npat);
		if (it!=PatMap.end()){
			// exist in cache
			patid_t patid = it->second;
//			dbw("%ws : num_vec=%d id=%d", nfilename.c_str(), PatInfoMap[nfilename].size(), patid);
			// patidɊY錟ʂT(linear search)
			//TODO: bsearchɂ
			foreach(PatInfoMap[nfilename], it, patinfovec_t){
				if (it->patid == patid){
					// found the pattern in the target file.
					if (it->option & SRCH_OPT_HIT){
						if (npatlen==orgnpatlen){
							// łɌς݂pattern
						} else {
							// new pattern
							return NULL;
						}
					} else {
						// łɌς݂pattern
						// non-hitłꍇ́AłpatternqbgȂ
					}
					return it;	// found
				}
			}
			// not exist the pattern in the target file
		}
		if (npatlen==orgnpatlen){
			npatlen--;
			npatbuf.set(npat, npatlen);
			npat = npatbuf;
		} else {
			npatlen--;
			npatbuf[npatlen] = '\0';
		}
	}
	
	return NULL;
}

void TFileCache::GC()
{
	// PatListdTAPatInfoMap[] array炻̕IDTA폜B
}

#include "prgcommon.h"

#define	CACHEDIR		_t("cache")
#define	TEMPCACHEDIR	_t("cache.tmp")
#define	PATTERNFILE		_t("pattern")
#define	FILEINFOFILE	_t("fileinfo")
#define	MAGIC_GEN		(2)						// magic generation
#define	MAGIC_CACHE		(0x38593829+MAGIC_GEN)	// magic number for cache file
#define	MAGIC_PATTERN	(0x73259234+MAGIC_GEN)	// magic number for pattern list file
#define	MAGIC_FILEINFO	(0x548e682a+MAGIC_GEN)	// magic number for file info file

// \/: -> _
void escapepath(tchar *s)
{
	for(;*s;){
		if (*s=='\\' || *s=='/' || *s==':')
			*s = '_';
		s++;
	}
}

bool TFileCache::Save(bool diff_save)
{
	tnstr path(CommandPath);
	if (diff_save){
		path += CACHEDIR;
		if (!fexist(path)){
			if (_tmkdir(path)!=0){
				return false;
			}
		}
	} else {
		path += TEMPCACHEDIR;
		DeleteFiles(path);
		if (fexist(path)){
			return false;
		}
		if (_tmkdir(path)!=0){
			return false;
		}
	}
	path += _t("\\");
	bool ret = true;
	Saving = true;

	tnstrimap_t savedfiles;
	ret = SavePatInfo(path, diff_save, savedfiles);
	
	if (ret){
		// save id-string list.
		ret = SavePattern(path);
	}
	Saving = false;

	if (!diff_save && ret){
		tnstr cachedir(CommandPath);
		cachedir += CACHEDIR;
		DeleteFiles(cachedir);	//TODO: ɎԂ
		if (!MoveFile(path, cachedir)){
			dbw("Cannot move %ws to %ws.", path.c_str(), cachedir.c_str());
			ret = false;
		}
	}
	// clear dirty flag
	if (ret){
		if (diff_save){
			foreach(savedfiles, it, tnstrimap_t){
				FileInfoMap[it->first].dirty = false;	// dirty flag cleared
			}
		}
		Modified = false;
	}
	return ret;
}

bool TFileCache::SavePatInfo(const tchar *path, bool diff_save, tnstrimap_t &savedfiles)
{
	bool ret = true;
	int pathlen = _tcslen(path);

	tnstrimap_t targetdirs;
	if (diff_save){
		// dirty filedirectory(->targetdirs)
		foreach(PatInfoMap, it, patinfomap_t){
			if (it->second.size()==0)
				continue;	// no data

			__assert(FileInfoMap.count(it->first));
			if (!FileInfoMap[it->first].dirty)
				continue;
			tchar *fname = GetFileName(it->first.c_str());
			tchar c = fname[0];
			fname[0] = '\0';
			targetdirs[it->first.c_str()] = true;
			fname[0] = c;
		}
	}
	
	typedef map< tnstr, FILE* > fds_t;
	fds_t fds;
	#define	MAX_FDS		16	// max open files.
	foreach(PatInfoMap, it, patinfomap_t){
		if (it->second.size()==0)
			continue;	// no data

		int is_target = 0;

		if (diff_save){
			tchar *fname = GetFileName(it->first.c_str());
			tchar c = fname[0];
			fname[0] = '\0';
			tnstrimap_t::iterator it2 = targetdirs.find(it->first.c_str());
			if (it2!=targetdirs.end()){
				is_target = it2->second;
			}
			fname[0] = c;
			if (!is_target)
				continue;
		}
		
		tnstrbuf name(path);
		tnstr infopath = _GetFilePath(it->first);
		if (infopath.empty())
			continue;
		const tchar *filename = GetFileName(it->first);
		//dbw("%ws : %d", it->first.c_str(), FileInfoMap[it->first].dirty);
		name += infopath;
		const tchar *savename = name.c_str()+pathlen;
		escapepath(name.c_str()+pathlen);	// \/ -> _

		ret = false;
		
		FILE *fp;
		unsigned short uslen;
		fds_t::iterator it_f = fds.find(savename);
		if (it_f!=fds.end()){
			// already opened
			fp = it_f->second;
		} else {
			if (fds.size()>=MAX_FDS){
				fclose(fds.begin()->second);
				//_tchmod(fds.begin()->first, S_IREAD|S_IWRITE);
				fds.erase(fds.begin());
			}
			if (is_target==1){
				fp = _tfopen(name, _t("wb"));
				__assert(targetdirs.count(infopath));
				targetdirs[infopath] = 2;	// mark as already created.
				//dbw("diff create:(%ws) %d", infopath.c_str(), targetdirs[infopath]);
			} else {
				fp = _tfopen(name, _t("ab"));
			}
			if (!fp){
				dbw("Cannot open %ws", name.c_str());
				break;
			}
			fds[savename] = fp;
			if (fseek(fp, 0, SEEK_CUR)==0){
				// top of the file.
				// write magic number
				int magic = MAGIC_CACHE;
				if (fwrite(&magic, sizeof(magic), 1, fp)<=0){
					break;
				}
				// write real file path
				uslen = (unsigned short)infopath.length();
				if (fwrite(&uslen, sizeof(uslen), 1, fp)<=0){
					break;
				}
				if (fwrite(infopath.c_str(), (int)uslen*sizeof(tchar), 1, fp)<=0){
					break;
				}
			}
		}
		if (diff_save){
			savedfiles[it->first] = true;
		}

		// filename char. count
		uslen = (unsigned short)_tcslen(filename);
		if (fwrite(&uslen, sizeof(uslen), 1, fp)<=0){
			break;
		}
		// real filename
		if (fwrite(filename, (int)uslen*sizeof(tchar), 1, fp)<=0){
			break;
		}
		// file timestamp hash
		if (fwrite(&FileInfoMap[it->first].tmhash, sizeof(tmhash_t), 1, fp)<=0){
			break;
		}
		// number of PatInfo elements
		const patinfovec_t &infovec = it->second;
		uslen = (unsigned short)infovec.size();
		if (fwrite(&uslen, sizeof(uslen), 1, fp)<=0){
			break;
		}
		// PatInfo vector.
		if (fwrite(infovec.begin(), sizeof(TPatInfo)*infovec.size(), 1, fp)<=0){
			break;
		}
		ret = true;
	}
	// close all files
	foreach(fds,it,fds_t){
		fclose(it->second);
		//_tchmod(it->first, S_IREAD|S_IWRITE);
	}

	return ret;
}

bool TFileCache::SavePattern(const tchar *path)
{
	bool ret = false;

	tnstrbuf name(path);
	name += PATTERNFILE;
	FILE *fp = _tfopen(name, _t("wb"));
	if (fp){
		int magic = MAGIC_PATTERN;
		if (fwrite(&magic, sizeof(magic), 1, fp)>0){
			ret = true;
			foreach(PatList, it, patlist_t){
				// patid
				if (fwrite(&it->first, sizeof(it->first), 1, fp)<=0){
					ret = false;
					break;
				}
				// length
				unsigned short len = (unsigned short)it->second.size();	// char. count
				if (fwrite(&len, sizeof(len), 1, fp)<=0){
					ret = false;
					break;
				}
				// string pattern
				if (fwrite(it->second.c_str(), it->second.size()*sizeof(tchar), 1, fp)<=0){	// without null term.
					ret = false;
					break;
				}
			}
		}
		fclose(fp);
	}
	return ret;
}

bool TFileCache::Load()
{
	tnstr path(CommandPath);
	path += CACHEDIR;
	if (!fexist(path)){
		return false;
	}

	path += _t("\\");

	bool ret = true;

	// Load pattern file.
	LoadPattern(path);

	// Load PatInfo
	LoadPatInfo(path);
	
	return ret;
}

bool TFileCache::LoadPattern(const tchar *path)
{
	bool ret = false;
	tnstr name(path);
	name += PATTERNFILE;
	FILE *fp = _tfopen(name, _t("rb"));
	if (fp){
		int magic;
		size_t r = fread(&magic, sizeof(magic), 1, fp);
		if (r==1 && magic==MAGIC_PATTERN){
			ret = true;
			tnstrbuf buf;
			for (;;){
				// patid
				patid_t patid;
				r = fread(&patid, sizeof(patid), 1, fp);
				if (r==0){
					break;
				}
				// length
				unsigned short len;
				r = fread(&len, sizeof(len), 1, fp);
				if (r==0){
					break;
				}
				// string pattern
				buf.expand(len+1);
				r = fread(buf.c_str(), (int)len*sizeof(tchar), 1, fp);
				if (r==0){
					break;
				}
				buf[len] = '\0';
				PatList[patid] = buf.c_str();
				PatMap[buf.c_str()] = patid;
				if (LastPatInfoId<patid)
					LastPatInfoId = patid;
			}
		}
		fclose(fp);
	}
	return ret;
}

#include <dirent.h>
bool TFileCache::LoadPatInfo(const tchar *path)
{
	_tDIR *dir = _topendir(path);
	if (!dir){
		return false;
	}
	_tdirent *ent;
	while ((ent=_treaddir(dir))!=NULL){
		tnstr name(path);
		name += ent->d_name;
		FILE *fp = _tfopen(name, _t("rb"));
		if (!fp)
			continue;
		int magic;
		size_t r = fread(&magic, sizeof(magic), 1, fp);
		if (r==1 && magic==MAGIC_CACHE){
			// read real path
			tnstrbuf filepath;
			tnstrbuf buf;
			unsigned short uslen;
			if (fread(&uslen, sizeof(uslen), 1, fp)!=1){
				goto jend;
			}
			filepath.expand(uslen);
			if (fread(filepath.c_str(), uslen*sizeof(tchar), 1, fp)!=1){
				goto jend;
			}
			filepath[uslen] = '\0';
			int filepath_len = uslen;
			for (;;){
				// filename char. count
				if (fread(&uslen, sizeof(uslen), 1, fp)!=1){
					break;
				}
				// real filename
				buf.expand(filepath_len+uslen);
				_tcscpy(buf.c_str(), filepath);
				tchar *namebuf = buf.c_str()+filepath_len;
				if (fread(namebuf, uslen*sizeof(tchar), 1, fp)!=1){
					break;
				}
				namebuf[uslen] = '\0';
				//DBW("file:%ws", buf.c_str());
				_NormalizeFileName(buf.c_str());
				// file timestamp hash
				tmhash_t tmhash;
				if (fread(&tmhash, sizeof(tmhash), 1, fp)!=1){
					break;
				}
				FileInfoMap_insert(buf.c_str(), tmhash, false);
				// number of PatInfo elements
				if (fread(&uslen, sizeof(uslen), 1, fp)!=1){
					break;
				}
				// PatInfo vector.
				for (int i=0;i<uslen;i++){
					TPatInfo info;
					if (fread(&info, sizeof(info), 1, fp)!=1){
						break;
					}
					//DBW("patid=%d option=%04X lasttime=%08X", info.patid, info.option, info.lasttime);
					PatInfoMap[buf.c_str()].push_back(info);
					if (CacheTime<info.lasttime)
						CacheTime = info.lasttime;
				}
			}
		}
jend:
		fclose(fp);
	}

	_tclosedir(dir);
	return true;
}

