//---------------------------------------------------------------------------

#ifndef FileCacheH
#define FileCacheH
//---------------------------------------------------------------------------

#include "defs.h"
#include "SrchDef.h"

typedef unsigned short patid_t;
typedef unsigned short srchopt_t;
typedef int tmhash_t; 	// time hash value
typedef int cachetm_t;	// cache time

struct TPatInfo {
	patid_t patid;		// [2] pattern id
	srchopt_t option;	// [2] search option and hit flag
	cachetm_t lasttime;	// [4] last search time

	TPatInfo(){}
	TPatInfo(patid_t _patid, srchopt_t _option, time_t _lasttime)
	{
		patid = _patid;
		option = _option;
		lasttime = _lasttime;
	}
};

struct TCFileInfo {
	tmhash_t tmhash;
	bool dirty;

	TCFileInfo()
	{
		tmhash = 0;
		dirty = false;
	}
	inline void assign(const TCFileInfo &o)
	{
		this->tmhash = o.tmhash;
		this->dirty = o.dirty;
	}
	inline void assign(tmhash_t _tmhash, bool _dirty)
	{
		this->tmhash = _tmhash;
		this->dirty = _dirty;
	}
	TCFileInfo(tmhash_t _tmhash, bool _dirty)
	{
		assign(_tmhash, _dirty);
	}
	TCFileInfo &operator = (const TCFileInfo &o)
	{
		assign(o);
		return *this;
	}
};

class TFileCache {
protected:
	typedef vector<TPatInfo> patinfovec_t;
	//TODO: typedef set<TPatInfo> patinfovec_t ̂قB
	typedef map< tnstr, patinfovec_t > patinfomap_t;
	typedef map< patid_t, tnstr > patlist_t;
	typedef map< tnstr, patid_t > patmap_t;
	//typedef map< tnstr, tmhash_t > filetmh_t;	// file timestamp hash
	typedef map< tnstr, TCFileInfo > fileinfomap_t;
	typedef map< tnstr, int > tnstrimap_t;

	patinfomap_t PatInfoMap;	// filename -> pattern info list map
	patlist_t PatList;	// pattern id -> string
	patmap_t PatMap;	// string -> pattern id
	patid_t LastPatInfoId;
	//filetmh_t FileTS;	// file timestamp hash map
	fileinfomap_t FileInfoMap;

	cachetm_t CacheTime;
	bool Modified;	//TODO: generation̂ق̂ł́H
	bool Saving;

	TMutex Mutex;
public:
	TFileCache();
	virtual ~TFileCache();
	void lock()
		{ Mutex.Lock(); }
	void unlock()
		{ Mutex.Unlock(); }
	void Clear();
protected:
	inline void AdvTime()
		{ CacheTime++; }
public:
	static tnstr NormalizePattern(const tchar *str);
	void Add(const tchar *filename, tmhash_t timehash, const tchar *srchpat, srchopt_t option, bool hit);
	void _Add(const tchar *filename, tnstr &nfilename, tmhash_t timehash, const tchar *nsrchpat, int nsrchpatlen, srchopt_t option, bool hit);
	int Find(const tchar *filename, tmhash_t timehash, const tchar *srchpat, srchopt_t option);
	int _Find(const tchar *filename, tnstr &nfilename, tmhash_t timehash, const tchar *nsrchpat, int nsrchpatlen, srchopt_t option);
	TPatInfo *__Find(const tchar *nfilename, const tchar *npat, int npatlen, srchopt_t option);
	void Erase(const tchar *filename);
	void __Erase(const tchar *nfilename);
	void GC();
	bool Save(bool diff_save=true);
	bool Load();
	bool IsModified() const
		{ return Modified; }
protected:
	TPatInfo *Find(const tchar *filename, tnstr &nfilename, const tchar *nsrchpat, int npatlen, srchopt_t option);
	bool SavePatInfo(const tchar *path, bool dirty_check, tnstrimap_t &savedfiles);
	bool SavePattern(const tchar *path);
	bool LoadPattern(const tchar *path);
	bool LoadPatInfo(const tchar *path);
};

#endif

