//---------------------------------------------------------------------------

#include "tnlib.h"
#pragma hdrstop
#include "verinfo.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)

TVersionInfo::TVersionInfo(const tchar *appFName)
{
	DWORD   fvHandle;

	FVData = NULL;

	tchar *appname = _tcsdup( appFName );
	DWORD dwSize = GetFileVersionInfoSize( appname, &fvHandle);
	if (dwSize) {
		FVData  = (void *)new tchar[dwSize];
		if (!GetFileVersionInfo( appname, fvHandle, dwSize, FVData)) {
			// error
			if (FVData){
				delete[] FVData;
				FVData = NULL;
			}
		}
	}
	free( appname );
}

TVersionInfo::~TVersionInfo()
{
	if (FVData)
		delete[] FVData;
}

bool TVersionInfo::GetProductVersion(tnstr &prodVersion)
{
	if (!FVData)
		return false;

	VS_FIXEDFILEINFO *pFixedFileInfo;
	UINT uLen;
	if(!VerQueryValue(FVData, _T("\\"), (LPVOID*)&pFixedFileInfo, &uLen)){
		return false;
	}
	prodVersion = tnsprintf(_t("%d.%d.%d"), 
		HIWORD(pFixedFileInfo->dwProductVersionMS),
		LOWORD(pFixedFileInfo->dwProductVersionMS),
		HIWORD(pFixedFileInfo->dwProductVersionLS)
		/*,LOWORD(pFixedFileInfo->dwProductVersionLS)*/);
	return true;
}

static const TCHAR *VERS_TRANSLATION = _T("\\varFileInfo\\translation");
static const TCHAR *VERS_STRINGFILEINFO = _T("\\StringFileInfo\\%08lx\\%s");
static const TCHAR *VERS_PRODUCTNAME = _T("ProductName");
//static const TCHAR *VERS_PRODUCTVERSION = _T("ProductVersion");
static const TCHAR *VERS_LEGALCOPYRIGHT = _T("LegalCopyright");
static const TCHAR *VERS_SPECIALBUILD = _T("SpecialBuild");

TVersionInfoString::TVersionInfoString(const tchar *appFName)
	:super(appFName)
{
	UINT vSize;

	if (FVData){
		// If not error in super class.
		tchar    subBlockName[255];
		_tcscpy(subBlockName, VERS_TRANSLATION);
		if (!VerQueryValue(FVData, subBlockName, (void **)&TransBlock, &vSize)) {
			delete[] FVData;
			FVData = NULL;
		} else {
			*(DWORD *)TransBlock = MAKELONG(HIWORD(*(DWORD *)TransBlock), LOWORD(*(DWORD *)TransBlock));
		}
	}
}

bool TVersionInfoString::GetProductName(tnstr &prodName)
{
	if (!FVData)
		return false;

	UINT    vSize;
	tchar    subBlockName[255];

	wsprintf(subBlockName, VERS_STRINGFILEINFO, *(DWORD *)TransBlock, VERS_PRODUCTNAME);
	LPTSTR pptr;
	if (!VerQueryValue(FVData, subBlockName, (void **)&pptr, &vSize)){
		return false;
	}
	prodName = pptr;
	return true;
}

bool TVersionInfoString::GetCopyright(tnstr &copyright)
{
	if (!FVData)
		return false;

	UINT    vSize;
	tchar    subBlockName[255];

	wsprintf(subBlockName, VERS_STRINGFILEINFO, *(DWORD *)TransBlock, VERS_LEGALCOPYRIGHT);
	LPTSTR pptr;
	if (!VerQueryValue(FVData, subBlockName, (void FAR* FAR*)&pptr, &vSize)){
		return false;
	}
	copyright = pptr;
	return true;
}


bool TVersionInfoString::GetDebug(tnstr &debug)
{
	if (!FVData)
		return false;

	UINT    vSize;
	tchar    subBlockName[255];

	wsprintf(subBlockName, VERS_STRINGFILEINFO, *(DWORD *)TransBlock, VERS_SPECIALBUILD);
	LPTSTR pptr;
	if (!VerQueryValue(FVData, subBlockName, (void FAR* FAR*)&pptr, &vSize)){
		return false;
	}
	debug = pptr;
	return true;
}

// Version comparison for application.
int CompareVersion(const tchar *v1, const tchar *v2)
{
	const int NUM_DIGIT	= 3;
	int n1[NUM_DIGIT] = {0,0,0};
	int n2[NUM_DIGIT] = {0,0,0};

	if(EOF == _stscanf(v1, _T("%d.%d.%d"), &n1[0], &n1[1], &n1[2])){
		// format error. assume to be zero.
	}
	if(EOF == _stscanf(v2, _T("%d.%d.%d"), &n2[0], &n2[1], &n2[2])){
		// format error. assume to be zero. 
	}

	for(int i=0;i<NUM_DIGIT;i++){
		if(n1[i] < n2[i]){
			return -1;	// n1 is older than n2.
		} else if(n1[i] == n2[i]){
			continue;
		} else {
			return 1;	// n1 is newer than n2.
		}
	}
	return 0;	// Same version.
}

