//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop
#include "tnlib.h"

#include "SquViewFrm.h"
#include "SrchProc.h"
#include "SquUIMain.h"
#include "SquIF.h"
#include "wcommon.h"
#include "MainCmd.h"
#include "prgstrlib.h"
#include "prgcommon.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "TntExtCtrls"
#pragma link "TntStdCtrls"
#pragma link "TntMenus"
#pragma resource "*.dfm"
TSquViewFrame *SquViewFrame;

#define	UM_DUMMY	(WM_APP+0xFFF)

bool GetFileInfo(const tchar *filename, SYSTEMTIME *st, unsigned __int64 *size, DWORD *attr);	//TODO: move to common

class TSearchView : public TSearchViewI {
protected:
	TSquViewFrame *View;
	TSquUIMain *Main;
	tnstr Msg;
	int SrchTime;
	int SrchCount;
	tnstr Warning;
	tnstr DispFileInfoName;
	tnstr FileInfo;
public:
	bool Searching;
protected:
	inline void _SetMsg(const tchar *str, int pasttime=-1, int count=-1)
		{ Main->SetMsg(View->squi(), str, pasttime, count); }
	inline void _SetWarning(const tchar *str)
		{ Main->SetWarning(View->squi(), str); }
	inline void _SetFileInfo(const tchar *str)
		{ Main->SetFileInfo(View->squi(), str); }
	void SetMsg()
	{
		_SetMsg(Msg, SrchTime, SrchCount);
		_SetWarning(Warning);
		_SetFileInfo(FileInfo);
	}
public:
	TSearchView(TSquViewFrame *view, TSquUIMain *main)
		:View(view)
		,Main(main)
	{
		Searching = false;
		SrchTime = -1;
		SrchCount = -1;
	}
	virtual void BeginUpdate()
	{
		View->lvResult->Items->BeginUpdate();
	}
	virtual void EndUpdate()
	{
		View->lvResult->Items->EndUpdate();
	}
	virtual void AddFile(const tchar *filename)
	{
		View->lvResult->Items->Add(filename);
	}
	virtual void AddFileLoc(int line, int col, const tchar *text)
	{
		//TODO: binary dataWideStringɓ삵Ȃ
		tnstr s = tnsprintf(_t("  %4d,%3d : "), line+1, col+1);
		s += text;
		View->lvResult->Items->Add(s);
	}
	virtual void UpdateStatus(int status)
	{
		switch (status){
			case ST_NONE:
				Searching = false;
				Msg.clear();
				SetMsg();
				break;
			case ST_SEARCHING:
				Msg = _t("Searching...");
				SetMsg();
				Searching = true;
				break;
		}
	}
	virtual void UpdateTime(int msec)
	{
		if (SrchTime!=msec){
			SrchTime = msec;
			SetMsg();
		}
	}
	// Note: called by another thread
	virtual void UpdateNotify()
	{
		// To kick the idle proc.
		if (View->LastIdleDone()){
			PostMessage(View->Handle, UM_DUMMY, 0, 0);
		}
	}
	// Note: called by another thread
	virtual void EndNotify()
	{
		// To kick the idle proc.
		if (View->LastIdleDone()){
			PostMessage(View->Handle, UM_DUMMY, 0, 0);
		}
	}
	void UpdateMsg()
	{
		SetMsg();
	}
	void SetWarning(const tchar *msg)
	{
		Warning = msg;
		_SetWarning(Warning);
	}
	void SetFileInfo(const tchar *filename)
	{
		if (!filename){
			if (DispFileInfoName.empty())
				return;
		} else
		if (DispFileInfoName==filename){
			return;
		}
		DispFileInfoName = filename;
		bool ok = false;
		if (DispFileInfoName.exist()){
			// Get file information //
			SYSTEMTIME st;
			uint64 size;
			if (GetFileInfo(filename, &st, &size, NULL)){
				tchar buf[40];
				FileInfo = tnsprintf(_t("%d/%d/%d %02d:%02d:%02d %s[bytes]"),
					st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond,
					itocs64(size, buf) );
				ok = true;
			}
		}
		if (!ok){
			FileInfo.clear();
		}
		_SetFileInfo(FileInfo);
	}
#if 0
	const tchar *GetMsg(int *pasttime)
	{
		if (pasttime)
			*pasttime = SrchTime;
		return Msg;
	}
#endif
};

//TODO: move to common
// Get file information
bool GetFileInfo(const tchar *filename, SYSTEMTIME *st, unsigned __int64 *size, DWORD *attr)
{
	HANDLE hFile = CreateFile(filename, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile==INVALID_HANDLE_VALUE)
		return false;

	bool ok = true;
	if (st){
		FILETIME ft,lft;
		if (!GetFileTime(hFile, NULL, NULL, &ft))
			ok = false;
		FileTimeToLocalFileTime(&ft, &lft);
		FileTimeToSystemTime(&lft, st);
	}
	if (ok && size){
		DWORD szHigh;
		DWORD szLow = GetFileSize(hFile, &szHigh);
		if (szLow==0xFFFFFFFF && GetLastError()!=NO_ERROR){
			ok = false;
		} else {
			*size = ((unsigned __int64)szLow) | (((unsigned __int64)szHigh)<<32);
		}
	}
	if (ok && attr){
		*attr = GetFileAttributes(filename);
	}
	CloseHandle(hFile);
	return true;
}


//---------------------------------------------------------------------------
__fastcall TSquViewFrame::TSquViewFrame(TComponent* Owner, TSquUIMain *main)
	: TTntFrame(Owner)
{
	SrchView = new TSearchView(this, main);
	SrchProc = new TSearchProcI(SrchView);
	SrchOption = 0;
	FLastIdleDone = false;
}
//---------------------------------------------------------------------------
void __fastcall TSquViewFrame::lvResultClick(TObject *Sender)
{
	SelChanged();
}
//---------------------------------------------------------------------------
void __fastcall TSquViewFrame::lvResultDblClick(TObject *Sender)
{
	KeyEnter();
}
//---------------------------------------------------------------------------
void __fastcall TSquViewFrame::lvResultDrawItem(TWinControl *Control,
	  int Index, TRect &Rect, TOwnerDrawState State)
{
	TCanvas &c = *((TTntListBox*)Control)->Canvas;
	const int MaxTextLen = 1024;
	tchar buf[MaxTextLen+1];

	int len;
	{	// Remove control chars.
		const tchar *sp = lvResult->Items->Strings[Index].c_bstr();
		lvResult->Items->Strings[Index].Length();
		try {	//TODO: binaryΉǂBB
			len = min(MaxTextLen, (int)_tcslen(sp));
		} catch(...){
			return;
		}
		tchar *dp = buf;
		const tchar *ep = sp + len;
		for (;sp<ep;){
			if (*(tuchar*)sp<' '){
				if (*(tuchar*)sp!='\r'){
					*dp++ = ' ';
				}
				sp++;
				continue;
			}
			*dp++ = *sp++;
		}
		*dp = '\0';
		len = STR_DIFF(dp, buf);	// adjust the length.
	}

	bool filename = !isdigit(buf[0]) && buf[0]!=' ';
	
	c.FillRect(Rect);
	if (filename){
		c.Font->Style = c.Font->Style << fsBold;
	} else {
		int cx = 2+Rect.left;
		const tchar *sp = buf;
		TSearchCommon sc(SrchPat, SrchOption);
		int srchpat_len = _tcslen(SrchPat);
		for(;;){
			if (!sc.Search(sp)){
				break;
			}

			// Draw normal and bold text.
			SIZE sz;
			if (sc.pos>0){
				TextOut(c.Handle, cx, Rect.top, sp, sc.pos);
				GetTextExtentPoint32(c.Handle, sp, sc.pos, &sz);
				sp += sc.pos;
				cx += sz.cx;
			}
			c.Font->Style = c.Font->Style << fsBold << fsUnderline;
			TextOut(c.Handle, cx, Rect.top, sp, srchpat_len);
			GetTextExtentPoint32(c.Handle, sp, srchpat_len, &sz);
			sp += srchpat_len;
			cx += sz.cx;
			c.Font->Style = c.Font->Style >> fsBold >> fsUnderline;
		}
		if (sp[0]){
			TextOut(c.Handle, cx, Rect.top, sp, _tcslen(sp));
		}
		return;
	}
	TextOut( c.Handle, 2+Rect.left, Rect.top, buf, len );
}
void __fastcall TSquViewFrame::cmOpenEditorClick(TObject *Sender)
{
	OpenEditor();
}
//---------------------------------------------------------------------------
void TSquViewFrame::Initialize()
{
}
void TSquViewFrame::Cleanup()
{
	if (SrchProc)
		delete SrchProc;
	if (SrchView)
		delete SrchView;
}
bool TSquViewFrame::IsSearching()
{
	return SrchView->Searching;
}

void TSquViewFrame::SelChanged()
{
	if (lvResult->ItemIndex>=0){
		tnstr filename;
		SrchProc->GetFile(lvResult->ItemIndex, filename);
		SrchView->SetFileInfo(filename);
	} else {
		SrchView->SetFileInfo(NULL);
	}
}

void TSquViewFrame::ResultSelDown()
{
	if (lvResult->ItemIndex == lvResult->Items->Count-1)
		return;	// bottom
	if (lvResult->ItemIndex<0){
		lvResult->ItemIndex = 0;
	} else {
		lvResult->ItemIndex = lvResult->ItemIndex+1;
	}
	_SelChanged();
}
void TSquViewFrame::ResultSelUp()
{
	if (lvResult->ItemIndex==0)
		return;	// top
	if (lvResult->ItemIndex<0){
		lvResult->ItemIndex = lvResult->Items->Count-1;
	} else {
		lvResult->ItemIndex = lvResult->ItemIndex-1;
	}
	_SelChanged();
}
void TSquViewFrame::NextSelFile()
{
	int i = lvResult->ItemIndex;
	if (i<0){
		ResultSelDown();
		return;
	}
	i = SrchProc->NextFileIndex(i);
	if (i>=0){
		lvResult->ItemIndex = i;
		_SelChanged();
	}
}
void TSquViewFrame::PrevSelFile()
{
	int i = lvResult->ItemIndex;
	if (i<0){
		ResultSelUp();
		return;
	}
	i = SrchProc->PrevFileIndex(i);
	if (i>=0){
		lvResult->ItemIndex = i;
		_SelChanged();
	}
}
void TSquViewFrame::CopyResult()
{
	if (lvResult->Items->Count==0)
		return;

	tnstrbuf buf;
	for (int i=0;i<lvResult->Items->Count;i++){
		buf += lvResult->Items->Strings[i].c_bstr();
		buf += _t("\n");
	}
	CopyClipboard(Handle, buf, buf.length());
}

void TSquViewFrame::KeyEnter()
{
	if (lvResult->ItemIndex>=0){
		OpenEditor();
	} else {
		lvResult->SetFocus();
	}
}
void TSquViewFrame::OpenEditor()
{
	int index = lvResult->ItemIndex;
	cmdOpenEditor(SrchProc, index);
}
void TSquViewFrame::OpenTool()
{
	int index = lvResult->ItemIndex;
	cmdOpenTool(SrchProc, index);
}
bool TSquViewFrame::StartSearch(const tchar *pat, const tchar *dir, const tchar *mask, int option)
{
	if (!dir || !dir[0] || !fexist(dir)){
		SrchView->SetWarning(_t("Bad directory!!"));
		return false;
	} else {
		SrchView->SetWarning(NULL);
	}
	if (SrchProc->StartSearch(pat, dir, mask, option)){
		lvResult->Items->Clear();
		SelChanged();
		SrchPat = pat;
		SrchDir = dir;
		SrchMask = mask;
		SrchOption = option;
		return true;
	} else {
	}
	return false;
}
void TSquViewFrame::GetSearchParams(struct TSearchParams &p)
{
	p.pat = SrchPat;
	p.dir = SrchDir;
	p.mask = SrchMask;
	p.option = SrchOption;
}
void TSquViewFrame::UpdateMsg()
{
	SrchView->UpdateMsg();
}
#if 0
const tchar *TSquViewFrame::GetMsg(int *pasttime)
{
	return SrchView->GetMsg(pasttime);
}
#endif
bool TSquViewFrame::IdleProc()
{
	return FLastIdleDone = SrchProc->IdleProc();
}


