//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop
#include "tnlib.h"
#include "profile.h"	// tnlib

//#include "pddefs.h"
#include "TntComCtrls.hpp"
#include "prgcommon.h"
//#include "wpdcom.h"

#include "vclcom.h"
//#include "vclfake.h"
//#include "txr.h"

//#include "tnfont.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)

#define	def_cstr(name, str)	\
	const tchar *name = _t(str)

def_cstr(sMain, "Main");
def_cstr(sLeft, "Left");
def_cstr(sTop, "Top");
def_cstr(sWidth, "Width");
def_cstr(sHeight, "Height");

def_cstr(sLVColWidth, "LVColWidth");

TShiftState sscNone, sscCtrl, sscShift, sscAlt;

class TInitializeVCL {
public:
	TInitializeVCL()
	{
		sscCtrl << ssCtrl;
		sscShift << ssShift;
		sscAlt << ssAlt;
	}
};

static TInitializeVCL ivcl;

void LoadForm(TRegKey *key, TForm *form, TWinControl *parent)
{
	LoadForm(key, form, parent?parent->Handle:NULL);
}

void LoadForm( TRegKey *key, TForm *form, HWND hwnd )
{
	RECT rc;
	if ( hwnd ){
		GetWindowRect( hwnd, &rc );
	} else {
		rc.left = rc.top = 0;
	}
	form->Left = key->ReadInteger(sLeft, form->Left) + rc.left;
	form->Top = key->ReadInteger(sTop, form->Top ) + rc.top;
	RECT rcScr;
	GetScreenSize(form->Handle?form->Handle:hwnd, &rcScr);
	int sx = rcScr.right - rcScr.left;
	int sy = rcScr.bottom - rcScr.top;

	switch ( form->BorderStyle ){
	case bsSizeable:
	case bsToolWindow:
	case bsSizeToolWin:
		form->Width = key->ReadInteger( sWidth, form->Width );
		form->Height = key->ReadInteger( sHeight, form->Height );
		break;
	}
	// ʂ̊OǂH
	if (form->Left+form->Width >= sx){
		// E[
		form->Left = sx - form->Width;
	}
	if (form->Left < rcScr.left){
		// [
		form->Left = rcScr.left;
	}
	if (form->Top+form->Height >= sy){
		// [
		form->Top = sy - form->Height;
	}
	if (form->Top < rcScr.top){
		// [
		form->Top = rcScr.top;
	}
}
void SaveForm( class TRegKey *key, class TForm *form, TWinControl *parent)
{
	SaveForm(key, form, parent?parent->Handle:NULL);
}
void SaveForm( TRegKey *key, TForm *form, HWND hwnd )
{
	RECT rc;
	if ( hwnd ){
		GetWindowRect( hwnd, &rc );
	} else {
		rc.left = rc.top = 0;
	}
	key->WriteInteger( sLeft, form->Left - rc.left );
	key->WriteInteger( sTop, form->Top - rc.top );
	switch ( form->BorderStyle ){
	case bsSizeable:
	case bsToolWindow:
	case bsSizeToolWin:
		key->WriteInteger( sWidth, form->Width );
		key->WriteInteger( sHeight, form->Height );
		break;
	}
}
void MoveCenter(TForm *form, TWinControl *parent)
{
	MoveCenter(form->Handle, parent ? parent->Handle : NULL);
}
void LoadListView(TRegKey *key, TTntListView *lv)
{
	// Load the width of columns of ListView
	for (int i=0;i<lv->Columns->Count;i++){
		TListColumn *col = lv->Columns->Items[i];
		tchar buf[20];
		wsprintf(buf, _T("%s_%d"), sLVColWidth, i);
		col->Width = key->ReadInteger(buf, col->Width);
	}
}
void SaveListView(TRegKey *key, TTntListView *lv)
{
	// Save the width of columns of ListView
	for (int i=0;i<lv->Columns->Count;i++){
		TListColumn *col = lv->Columns->Items[i];
		tchar buf[20];
		wsprintf(buf, _T("%s_%d"), sLVColWidth, i);
		key->WriteInteger(buf, col->Width);
	}
}
//---------------------------------------------------------------------------
//	Mouse operations
//---------------------------------------------------------------------------
int ShiftStatusToMouseKey(TShiftState shift)
{
	int ret = 0;
	if (shift.Contains(ssShift)){
		ret |= MK_SHIFT;
	} else
	if (shift.Contains(ssAlt)){
//		ret |= MK_MENU;	//TODO: not supported
	} else
	if (shift.Contains(ssCtrl)){
		ret |= MK_CONTROL;
	} else
	if (shift.Contains(ssLeft)){
		ret |= MK_LBUTTON;
	} else
	if (shift.Contains(ssRight)){
		ret |= MK_RBUTTON;
	} else
	if (shift.Contains(ssMiddle)){
		ret |= MK_MBUTTON;
	}
	return ret;
}
int ShiftStatusToKeyState(TShiftState shift)
{
	int ret = 0;
	if (shift.Contains(ssShift)){
		ret |= SK_SHIFT;
	}
	if (shift.Contains(ssAlt)){
		ret |= SK_ALT|SK_ALTMASK;
	}
	if (shift.Contains(ssCtrl)){
		ret |= SK_CTRL;
	}
	return ret;
}

//---------------------------------------------------------------------------
//	Data transfer
//---------------------------------------------------------------------------
void DT(TransferDir dir, TTntRadioGroup *rg, int &val)
{
	if (dir==tdSet)
		rg->ItemIndex = (int)val;
	else
		val = rg->ItemIndex;
}
void DT(TransferDir dir, TTntComboBox *cb, int &val)
{
	if (dir==tdSet)
		cb->ItemIndex = (int)val;
	else
		val = cb->ItemIndex;
}
void SetData(TTntComboBox *cb, HistoryArray *array)
{
	cb->Items->Clear();
	for ( int i=array->GetCount()-1;i>=0;i-- ){
		cb->Items->Add( _toString((*array)[i].string) );
	}
	if ( array->GetCount() ){
		cb->ItemIndex = 0;
		cb->Text = _toString((*array)[array->GetCount()-1].string);
	} else {
		cb->Text = _T("");
	}
}
void GetData(TTntComboBox *cb, HistoryArray *array)
{
	array->Clear();
	for ( int i=cb->Items->Count-1;i>=0;i-- ){
		array->AddString(cb->Items->Strings[i]);
	}
	array->AddString(cb->Text);
}
void DT(TransferDir dir, TTntComboBox *cb, HistoryArray *array)
{
	if (dir==tdSet){
		cb->Text = "";
		cb->Items->Clear();
		for ( int i=array->GetCount()-1;i>=0;i-- ){
			cb->Items->Add( _toString((*array)[i].string) );
		}
		if ( array->GetCount() ){
			cb->ItemIndex = 0;
		}
	} else {
		array->Clear();
		for ( int i=cb->Items->Count-1;i>=0;i-- ){
			array->AddString( cb->Items->Strings[i] );
		}
		array->AddString( cb->Text );
	}
}
#if defined(__HGRPARR_H)
void DT(TransferDir dir, TTntComboBox *cb, class HistoryGroupArray *array)
{
	if (array->GetSel()==-1){
		DT(dir, cb, (HistoryArray*)array);
		return;
	}
	if (dir==tdSet){
		// O[vƂ
		cb->Items->Clear();
		for ( int i=0;i<array->GetCount();i++ ){
			cb->Items->Add( _toString((*array)[i].string) );
		}
		if ( array->GetCount() ){
			cb->ItemIndex = array->GetSel( );
		}
	} else {
		// O[vƂ
		array->SetSel( cb->ItemIndex );
	}
}
#endif
void DT(TransferDir dir, TTntCheckBox *cb, bool &checked)
{
	if (dir==tdSet){
		cb->Checked = checked;
	} else {
		checked = cb->Checked;
	}
}
void DT(TransferDir dir, TTntCheckBox *cb, int &val, int flag)
{
	if (dir==tdSet){
		cb->Checked = (val & flag)?true:false;
	} else {
		val = (cb->Checked ? flag:0) | (val & ~flag);
	}
}
void DT(TransferDir dir, TTntRadioButton *rb, bool &checked)
{
	if (dir==tdSet){
		rb->Checked = checked;
	} else {
		checked = rb->Checked;
	}
}

void DT(TransferDir dir, TTntEdit *ed, int &val)
{
	if (dir==tdSet){
		tchar buf[40];
		_itot(val, buf, 10);
		ed->Text = buf;
	} else {
		val = _ttoi(_totchar(ed->Text));
	}
}
void DT(TransferDir dir, TTntLabel *label, tnstr &text)
{
	if (dir==tdSet){
		label->Caption = text;
	} else {
		text = label->Caption;
	}
}
void DT(TransferDir dir, TTntUpDown *ud, int &val)
{
	if (dir==tdSet){
		ud->Position = val;
	} else {
		val = ud->Position;
	}
}

void DT(TransferDir dir, TTntTrackBar *ud, int &val)
{
	if (dir==tdSet){
		ud->Position = val;
	} else {
		val = ud->Position;
	}
}
void DT(TransferDir dir, TRadioGrouping &rg, int &index)
{
	if (dir==tdSet){
		rg.SetItemIndex(index);
	} else {
		index = rg.GetItemIndex();
	}
}

void SetItems(TTntComboBox *cb, const tnstr_vec &array)
{
	SetItems(cb->Items, array);
}
void SetItems(TTntListBox *cb, const tnstr_vec &array)
{
	SetItems(cb->Items, array);
}
void SetItems(TTntStrings *strs, const tnstr_vec &array)
{
	strs->Clear();
	for (int i=0;i<array.get_num();i++){
		strs->Add(array[i]);
	}
}
#if defined(txrH)
void SetItems(TTntComboBox *cb, int idlist[])
{
	for (int i=0;idlist[i];i++){
		cb->Items->Add(_LT(idlist[i]));
	}
}
#endif
#if 0
void SetItems(TTntComboBox *cb, int start_id, int end_id)
{
	for (;start_id<=end_id;start_id++){
		cb->Items->Add(_LT(start_id));
	}
}
#endif
void SetItems(TTntComboBox *cb, const tchar *strings, const tchar *delim)
{
	if (!delim)
		delim = _t("\t");

	tchar *buf = _tcsdup(strings);
	if (!buf)
		return;	// memory error
	tchar *p = _tcstok(buf, delim);
	while (p){
		cb->Items->Add(p);
		p = _tcstok(NULL, delim);
	}
	free(buf);
}

void GetItems(TTntListBox *lb, tnstr_vec &array)
{
	GetItems(lb->Items, array);
}

void GetItems(TTntStrings *strs, tnstr_vec &array)
{
	array.clear();
	for (int i=0;i<strs->Count;i++){
		array.add(strs->Strings[i]);
	}
}

//---------------------------------------------------------------------------
//	Font
//---------------------------------------------------------------------------
int ToInteger(TFontStyles fs)
{
	int ret = 0;
	if (fs.Contains(fsBold))
		ret |= 1;
	if (fs.Contains(fsItalic))
		ret |= 2;
	if (fs.Contains(fsUnderline))
		ret |= 4;
	if (fs.Contains(fsStrikeOut))
		ret |= 8;
	return ret;
}
TFontStyles ToFontStyles(int val)
{
	TFontStyles fs;
	if (val & 1)
		fs << fsBold;
	if (val & 2)
		fs << fsItalic;
	if (val & 4)
		fs << fsUnderline;
	if (val & 8)
		fs << fsStrikeOut;
	return fs;
}

int VclColorToSysColor(TColor color)
{
	if (color==clWindow){
		return GetSysColor( COLOR_WINDOW );
	} else
	if (color==clWindowText){
		return GetSysColor( COLOR_WINDOWTEXT );
	}

	__assert(color<0x1000000);
	return color;
}
TColor SysColorToVclColor(COLORREF color, TColor defcolor)
{
	if (color==(COLORREF)-1)
		return defcolor;
	return (TColor)color;
}
void SetFont(TFont *font, LOGFONT &lf, bool useNormalSize)
{
	font->Name = lf.lfFaceName;
	if (useNormalSize){
		font->Height = 0;	// Uses default size.
	} else {
		font->Height = lf.lfHeight;
	}
	font->Charset = lf.lfCharSet;
	switch (lf.lfPitchAndFamily){
		case FIXED_PITCH:
			font->Pitch << fpFixed;
			break;
		case VARIABLE_PITCH:
			font->Pitch << fpVariable;
			break;
	}
	//font->Size = 
	TFontStyles fs = ExtractFontStyle(lf);
	font->Style = fs;
}
// Return the proper height to show texts.
int SetFont(TFont *font, LOGFONT &lf, bool useNormalSize, HDC hdc, bool includeAscent)
{
	::SetFont(font, lf, useNormalSize);

	if (hdc){
		//TODO: K؂ȍ߂@𒲂ׂ
		int lfHeight = lf.lfHeight;	// backup
		if (useNormalSize)
			lf.lfHeight = 0;
		HFONT hFont = CreateFontIndirect(&lf);
		lf.lfHeight = lfHeight;
		if (hFont){
			HFONT hOldFont = (HFONT)SelectObject(hdc, hFont);
			TEXTMETRIC tm;
			GetTextMetrics(hdc, &tm);
			SelectObject(hdc, hOldFont);
			DeleteObject(hFont);
			return tm.tmHeight + tm.tmDescent + (includeAscent?tm.tmAscent:0);
		} else {
			return 0;
		}
	} else {
		return font->Size;
	}
}

TFontStyles ExtractFontStyle(const LOGFONT &lf)
{
	TFontStyles fs;
	switch (lf.lfWeight){
		case FW_BOLD:
			fs << fsBold;
			break;
		default:
			fs >> fsBold;
			break;
	}
	if (lf.lfItalic){
		fs << fsItalic;
	}
	if (lf.lfUnderline){
		fs << fsUnderline;
	}
	if (lf.lfStrikeOut){
		fs << fsStrikeOut;
	}
	return fs;
}
#if 0	// temporary out
void GetFont(TFont *font, LOGFONT &lf)
{
	memset(&lf, 0, sizeof(lf));
	_tcsncpy(lf.lfFaceName, _mt(font->Name.c_str()), sizeof(lf.lfFaceName)/sizeof(tchar)-1);
	lf.lfHeight = font->Height;
	lf.lfCharSet = font->Charset;
	if (font->Pitch == fpFixed){
		lf.lfPitchAndFamily = FIXED_PITCH;
	} else
	if (font->Pitch == fpVariable){
		lf.lfPitchAndFamily = VARIABLE_PITCH;
	}
	GetFontStyle(font->Style, lf);
}
void GetFontStyle(const TFontStyles &style, LOGFONT &lf)
{
	if (style.Contains(fsBold)){
		lf.lfWeight = FW_BOLD;
	}
	if (style.Contains(fsItalic)){
		lf.lfItalic = TRUE;
	}
}

void SetFont(TFont *font, const class TFontAttr &fa, bool useNormalSize)
{
	LOGFONT lf;
	fa.Get(lf);
	SetFont(font, lf, useNormalSize);
}
int SetFont(TFont *font, const TFontAttr &fa, bool useNormalSize, HDC hdc, bool includeAscent)
{
	LOGFONT lf;
	fa.Get(lf);
	return SetFont(font, lf, useNormalSize, hdc, includeAscent);
}
TFontStyles ExtractFontStyle(const TFontAttr &fa)
{
	TFontStyles fs;
	if (fa.IsBold()){
		fs << fsBold;
	}
	if (fa.IsItalic()){
		fs << fsItalic;
	}
	if (fa.IsUnderline()){
		fs << fsUnderline;
	}
	if (fa.IsStrikeOut()){
		fs << fsStrikeOut;
	}
	return fs;
}
void GetFont(TFont *font, TFontAttr &fa)
{
	LOGFONT lf;
	GetFont(font, lf);
	fa.Set(lf);
}
void GetFontStyle(const TFontStyles &style, TFontAttr &fa)
{
	if (style.Contains(fsBold)){
		fa.SetBold(true);
	}
	if (style.Contains(fsItalic)){
		fa.SetItalic(true);
	}
}

//---------------------------------------------------------------------------
//	Memo
//---------------------------------------------------------------------------
//Note:
// Uvectorɂ̂őȂB
// ɂ́ATStreamhĂKvB
#if 0
// string reference stream
class TStrStream : public TStream {
protected:
	const tchar *Str;
	int StrSize;	// in character
public:
	TStrStream(const tchar *orgptr, int size)
		:Str(orgptr)
		,StrSize(size)
	{}
	// Count is char unit??
	virtual int __fastcall Read(void *Buffer, int Count);
};
//TOOD: mFĂȂ
int __fastcall TStrStream::Read(void *Buffer, int Count)
{
	void *org_buffer = Buffer;
	int size = StrSize<Count?StrSize:Count;
	int len = size;
	for (;len>0;len--){
		tchar c = *Str++;
		if (!c){
			Str--;
			break;
		}
		if (c=='\r')
			continue;
		if (c=='\n')
			break;
		*(tchar*)Buffer = c;
		Buffer = (tchar*)Buffer + 1;
	}
	*(tchar*)Buffer = '\0';
	StrSize -= (size-len);
	return (int)((char*)Buffer-(char*)org_buffer);
}
#endif
void vclSetLines(TTntMemo *edit, const tchar *text)
{
	tnstr_vec lines;
	lines.split(text, _t("\n"));
	edit->Lines->Clear();
	foreach_tnstr_vec(lines, it){
		edit->Lines->Add(*it);
	}
}

//---------------------------------------------------------------------------
//	UpDown
//---------------------------------------------------------------------------
// controlĂƂ`FbNȂ̂Œ
bool vclCanCloseUpDown(TTntUpDown *ud)
{
	if (!ud->Associate
		|| !ud->Enabled
		|| !IsWindowVisible(ud->Handle)){
		return true;	// does't matter
	}

	TTntEdit *edit = dynamic_cast<TTntEdit*>(ud->Associate);
	if (!edit)
		return true;	// I don't know.

	int val = _ttoi(edit->Text.c_bstr());
	return val>=ud->Min && val<=ud->Max;
}
//---------------------------------------------------------------------------
//	TRadioGrouping
//---------------------------------------------------------------------------
#define	foreach(obj, it, type) \
	for (type::iterator it=(obj).begin();it!=(obj).end();it++)
TRadioGrouping::TRadioGrouping()
{
}
TRadioGrouping::~TRadioGrouping()
{
}
int TRadioGrouping::GetItemIndex()
{
	for (int i=0;i<(int)Buttons.size();i++){
		TRadioButton *btn = Buttons.at(i);
		if (btn->Checked)
			return i;
	}
	return -1;
}
void TRadioGrouping::SetItemIndex(int index)
{
	for (int i=0;i<(int)Buttons.size();i++){
		TRadioButton *btn = Buttons.at(i);
		btn->Checked = (i==index);
	}
}

//---------------------------------------------------------------------------
//	Cursor
//---------------------------------------------------------------------------
void SetCursor(TControl *ctrl, const tchar *name)
{
	enum TMyCursor { crHand=1 };
	switch ((int)name){
		case 0:
			ctrl->Cursor = crDefault;
			break;
		case IDC_ARROW:
			ctrl->Cursor = crArrow;
			break;
		case IDC_HAND:
			//Cursor = crHandPoint;
			Screen->Cursors[crHand] = LoadCursor(0, IDC_HAND);
			ctrl->Cursor = (TCursor)crHand;
			break;
		default:
			__assert_debug;	// no cursor defined.
			break;
	}
}

//---------------------------------------------------------------------------
//	Dialog
//---------------------------------------------------------------------------
// Modeless dialogɂmodal dialog
void ShowDialog(TTntForm *dlg, TWinControl *parent)
{
	if (parent)
		parent->Enabled = false;
	if (dlg)
		dlg->Show();
}
void CloseDialog(TTntForm *dlg, TWinControl *parent)
{
	if (parent)
		parent->Enabled = true;
	DlgMessageProc();
	if (parent)
		parent->SetFocus();
	if (dlg)
		dlg->Close();
}
// \Ădlgclose邾B
// ʂdlgɐ؂ւꍇɎgp
void ChangeCloseDialog(TTntForm *dlg, TWinControl *parent)
{
	if (parent)
		parent->Enabled = true;
	if (dlg)
		dlg->Close();
}

class TWinControlRGDummy : public TWinControl {
typedef TWinControl super;
public:
	TWinControl* __fastcall FindNextControl(TWinControl* CurControl, bool GoForward, bool CheckTabStop, 
		bool CheckParent)
		{ return super::FindNextControl(CurControl, GoForward, CheckTabStop, CheckParent); }
};

// VCL bug remedy.
// refer to:
// http://www.asahi-net.or.jp/~HA3T-NKMR/tips014.htm
// TabړɂfocusOKAShift+Tabɂړ
void VCLRadioGroupFocusHandler(TTntForm *form, TObject *sender, bool enter)
{
	TWinControl *ActiveControl = form->ActiveControl;
	if (enter){
		//DBW("Focus in");
		// Focus in the radio group.
		if (ActiveControl==sender){
			TWinControlRGDummy *p = (TWinControlRGDummy*)ActiveControl->Parent;
			if (p){
				form->ActiveControl = p->FindNextControl(ActiveControl, true, true, true);
			}
		}
	} else {
		// Focus out from the radio group.
		//TODO: Ή - eventǂĎ󂯎邩H
		//DBW("Focus out");
		if (ActiveControl==sender){
			TWinControlRGDummy *p = (TWinControlRGDummy*)ActiveControl->Parent;
			if (p){
				form->ActiveControl = p->FindNextControl(ActiveControl, true, true, true);
			}
		}
	}
}

// Radio GroupɊmfocusZbg
//TODO: debugging now...
void VCLRadioGroupSetFocus(TTntForm *form, TTntRadioGroup *rg)
{
	rg->SetFocus();
	TWinControlRGDummy *p = (TWinControlRGDummy*)rg->Parent;
	if (p){
		form->ActiveControl = p->FindNextControl(rg, true, true, true);
	}
}

// SetFocus bug //
//#include "VclBugWin.h"

//TODO: VCL or TntVCLbugɂATTntRadioGroup handler̂ƂSetFocus()IME𖳌ɂĂ܂
// ܂ڍׂׂ͒ĂȂB::SetFocus()ł_ȂƂfocusړɌ肪ƎvB
// ƎvDummyWindowAŏ悤ƎvAvclcomlibformǉƁA
// exelink.dfmG[BB
// Ǒ̊windowԎ؂肷邱ƂɁB
enum {
	VCLBUG_SETFOCUS,
};
static HWND hwndVclBugHelper = NULL;
static int msgVclBugHelper;
void vclSetVclBugHelper(HWND hwnd, int msg)
{
	hwndVclBugHelper = hwnd;
	msgVclBugHelper = msg;
}
void vclVclBugHandler(WPARAM wparam, LPARAM lparam)
{
	switch (wparam){
		case VCLBUG_SETFOCUS:
			::SetFocus((HWND)lparam);
			break;
	}
}
void vclSetFocus(HWND handle)
{
	__assert(hwndVclBugHelper);
	PostMessage(hwndVclBugHelper, msgVclBugHelper, VCLBUG_SETFOCUS, (LPARAM)handle);
#if 0
	TVclBugWindow *form = TVclBugWindow::GetInstance();
	if (!form)
		return;
	form->SetFocus(handle);
#endif
}

void vclProcessMessage()
{
	vclProcessMessage(Application);
}

void vclHandleMessage()
{
	Application->HandleMessage();
}
#endif	// 0

void ListItemExchange(TTntListView *lv, int i1, int i2)
{
	TTntListItem *item1 = lv->Items->Item[i1];
	TTntListItem *item2 = lv->Items->Item[i2];
	TTntListItem *item = lv->Items->Add();
	item->Assign(item1);
	item1->Assign(item2);
	item2->Assign(item);
	item->Delete();
}



